import torch
import torchvision.utils as tv_utils
from torch import nn

import meddlr.ops as oF
from meddlr.config.config import configurable
from meddlr.modeling.meta_arch.build import META_ARCH_REGISTRY, build_model
from meddlr.ops import complex as cplx
from meddlr.transforms.base.mask import KspaceMaskTransform
from meddlr.transforms.gen.mask import RandomKspaceMask
from meddlr.utils.events import get_event_storage
from meddlr.utils.general import move_to_device
from meddlr.utils.transforms import SenseModel


@META_ARCH_REGISTRY.register()
class SSDUModel(nn.Module):
    """Self-supervised learning via data undersampling.

    This model is the relaxed form of the SSDU model that can be
    used to train with both supervised and unsupervised data.

    The mask used to acquire the data (:math:`\Omega`) is partitioned
    into train mask for the zero-filled image (:math:`\Theta`) and a
    mask for the loss (:math:`\Lambda`).

    Reference:
        B Yaman, SAH Hosseini, S Moeller, et al. Self-supervised
        learning of physics-guided reconstruction neural networks
        without fully sampled reference data.
        https://onlinelibrary-wiley-com.stanford.idm.oclc.org/doi/full/10.1002/mrm.28378
    """

    _version = 1

    @configurable
    def __init__(self, model: nn.Module, masker: RandomKspaceMask, vis_period: int = None):
        """
        Args:
            model (nn.Module): The base model.
            masker (NoiseModel): The additive noise module.
        """
        super().__init__()
        self.model = model
        self.masker = masker
        # Visualization done by this model
        if hasattr(self.model, "vis_period"):
            if vis_period is not None:
                self.model.vis_period = vis_period
            else:
                vis_period = self.model.vis_period
        self.vis_period = vis_period

    def augment(self, inputs):
        """Noise augmentation module for the consistency branch.

        Args:
            inputs (Dict[str, Any]): The input dictionary.
                It must contain a key ``'kspace'``, which traditionally
                corresponds to the undersampled kspace when performing
                augmentation for consistency. For supervised examples,
                this can correspond to either the retrospectively
                undersampled k-space or the fully-sampled kspace.

        Returns:
            Dict[str, Any]: The input dictionary with the kspace polluted
                with additive masked complex Gaussian noise.
        """
        masker = self.masker
        kspace = inputs["kspace"].clone()
        mask = cplx.get_mask(kspace)

        tfm: KspaceMaskTransform = masker.get_transform(kspace)
        train_mask = tfm.generate_mask(kspace, channels_last=True)
        loss_mask = mask - train_mask
        # TODO (arjundd): See if we can remove this check for speed reasons.
        assert torch.all(loss_mask >= 0)

        inputs = {k: v.clone() for k, v in inputs.items() if k != "kspace"}
        inputs["kspace"] = train_mask * kspace
        return inputs, mask[..., 0:1], train_mask, loss_mask[..., 0:1]

    @torch.no_grad()
    def visualize(self, images_dict):
        for name, images in images_dict.items():
            storage = get_event_storage()
            if isinstance(images, (tuple, list)):
                images = torch.stack(images, dim=0)
            if cplx.is_complex_as_real(images) or cplx.is_complex(images):
                images = {
                    f"{name}-phase": cplx.angle(images),
                    f"{name}-mag": cplx.abs(images),
                }
            else:
                images = {name: images}

            for name, data in images.items():
                if data.shape[-1] == 1:
                    data = data.squeeze(-1)
                data = data.unsqueeze(1)
                data = tv_utils.make_grid(
                    data, nrow=len(data), padding=1, normalize=True, scale_each=True
                )
                storage.put_image("ssdu/{}".format(name), data.cpu().numpy(), data_format="CHW")

    def forward(self, inputs):
        if not self.training:
            assert (
                "unsupervised" not in inputs
            ), "unsupervised inputs should not be provided in eval mode"
            inputs = inputs.get("supervised", inputs)
            return self.model(inputs)

        # Put supervised and unsupervised scans in a single tensor.
        sup = inputs.get("supervised", {})
        unsup = inputs.get("unsupervised", {})
        if sup or unsup:
            inputs = {
                k: torch.cat([sup.get(k, torch.tensor([])), unsup.get(k, torch.tensor([]))])
                for k in sup.keys() | unsup.keys()
            }
        assert all(k in inputs for k in ["kspace"])

        device = next(self.model.parameters()).device
        inputs = move_to_device(inputs, device=device, non_blocking=True)

        kspace = inputs["kspace"]
        inputs_aug, orig_mask, train_mask, loss_mask = self.augment(inputs)
        outputs = self.model(inputs_aug)

        # Get the signal model reconstructed images.
        # TODO: Make it possible to use these are the target instead of multi-coil images.
        pred_img, target_img, zf_image = outputs["pred"], outputs["target"], outputs["zf_image"]

        # Use signal model (SENSE) to get weighted kspace.
        A = SenseModel(maps=inputs_aug["maps"])  # no weights - we do not want to mask the data.
        loss_pred_kspace = loss_mask * A(outputs["pred"], adjoint=False)
        loss_kspace = loss_mask * kspace

        # A hacky way to prepare the predictions and target for the loss.
        # This may result in inaccurate training metrics outside of the loss.
        # TODO (arjundd): Fix this.
        outputs["pred"] = oF.ifft2c(loss_pred_kspace, channels_last=True)
        outputs["target"] = oF.ifft2c(loss_kspace, channels_last=True)

        # Visualize.
        if self.training and self.vis_period > 0:
            with torch.no_grad():
                storage = get_event_storage()
                if storage.iter % self.vis_period == 0:
                    A = SenseModel(maps=inputs["maps"][0:1], weights=train_mask[0:1])
                    base_image = A(kspace[0:1], adjoint=True)
                    self.visualize(
                        {
                            "masks": [orig_mask[0], train_mask[0], loss_mask[0]],
                            "kspace": [
                                kspace[0, ..., 0:1],
                                inputs_aug["kspace"][0, ..., 0:1],
                                loss_pred_kspace[0, ..., 0:1],
                                loss_kspace[0, ..., 0:1],
                            ],
                            "images": [
                                x[0]
                                for x in [base_image, zf_image, pred_img, target_img]
                                if x is not None
                            ],
                        }
                    )

        return outputs

    @classmethod
    def from_config(cls, cfg):
        model_cfg = cfg.clone()
        model_cfg.defrost()
        model_cfg.MODEL.META_ARCHITECTURE = cfg.MODEL.SSDU.META_ARCHITECTURE
        model_cfg.freeze()
        model = build_model(model_cfg)

        # TODO: Configure this
        params = cfg.MODEL.SSDU.MASKER.PARAMS
        masker = RandomKspaceMask(**params)
        masker.to(cfg.MODEL.DEVICE)

        return {"model": model, "masker": masker}



from meddlr.ops import categorical, complex, fft, utils  # noqa: F401
from meddlr.ops.categorical import (  # noqa: F401
    categorical_to_one_hot,
    logits_to_prob,
    one_hot_to_categorical,
    pred_to_categorical,
)
from meddlr.ops.fft import (  # noqa: F401
    fft2c,
    fft3c,
    fftc,
    fftnc,
    fftshift,
    ifft2c,
    ifft3c,
    ifftc,
    ifftnc,
    ifftshift,
)
from meddlr.ops.utils import (  # noqa: F401
    center_crop,
    normalize,
    normalize_instance,
    pad,
    roll,
    sliding_window,
    time_average,
    zero_pad,
)

# Do not add complex utilities to module level import.
# Complex utilites do not support real numbers
# and can cause clashes in the namespace.
__all__ = []
__all__.extend(fft.__all__)
__all__.extend(utils.__all__)
__all__.extend(categorical.__all__)



from typing import Sequence, Union

import torch

from meddlr.ops import complex as cplx
from meddlr.utils.events import get_event_storage


class NoiseModel:
    """A model that adds additive white noise.

    This model adds zero-mean Gaussian noise to a real-valued or
    complex-valued input. The standard deviation of the noise
    distribution is chosen from a range ``std_devs`` provided by
    the user. The standard deviation is often referred to as the
    *difficulty*, as higher standard deviations correspond to
    wider distributions from which noise can be sampled.

    To simulate noise during MRI acquisition, we only add noise
    to the samples that are acquired. In other words, if a sample
    in kspace was not acquired (i.e. accelerated acquisition), those
    samples are not corrupted with noise.

    Attributes:
        std_devs (Tuple[float]): The range of standard deviations used for
            noise model. If a single value is provided on initialization,
            it is stored as a tuple with 1 element.
        warmup_method (str): The method that is being used for warmup.
        warmup_iters (int): Number of iterations to use for warmup.
        generator (torch.Generator): The generator that should be used for all
            random logic in this class.

    Note:
        This class is functionally deprecated and will not be maintained.
        Use :cls:`RandomNoise` instead.

    Note:
        We do not store this as a module or else it would be saved to the model
        definition, which we dont want.

    Note:
        There is a known bug that the warmup method does not clamp the upper
        bound at the appropriate value. Thus the upper bound of the range keeps
        growing. We have elected not to correct for this to preserve
        reproducibility for older results. To use schedulers with corrected
        functionality, see :cls:`RandomNoise` instead.
    """

    def __init__(
        self,
        std_devs: Union[float, Sequence[float]],
        scheduler=None,
        mask=None,
        seed=None,
        device=None,
    ):
        """
        Args:
            std_devs (float, Tuple[float, float]): The  noise difficulty range.
            scheduler (CfgNode, optional): Config detailing scheduler.
            mask (CfgNode, optional): Config for masking method. Should have
                an attribute ``RHO`` that specifies the extent of masking.
            seed (int, optional): A seed for reproducibility.
        """
        if not isinstance(std_devs, Sequence):
            std_devs = (std_devs,)
        elif len(std_devs) > 2:
            raise ValueError("`std_devs` must have 2 or fewer values")
        self.std_devs = std_devs

        self.warmup_method = None
        self.warmup_iters = 0
        if scheduler is not None:
            self.warmup_method = scheduler.WARMUP_METHOD
            self.warmup_iters = scheduler.WARMUP_ITERS

        # Amount of the kspace to augment with noise.
        self.rho = None
        if mask is not None:
            self.rho = mask.RHO

        # For reproducibility.
        g = torch.Generator(device=device)
        if seed:
            g = g.manual_seed(seed)
        self.generator = g

    def choose_std_dev(self):
        """Chooses a random acceleration rate given a range."""
        if not isinstance(self.std_devs, Sequence):
            return self.std_devs
        elif len(self.std_devs) == 1:
            return self.std_devs[0]

        if self.warmup_method:
            curr_iter = get_event_storage().iter
            warmup_iters = self.warmup_iters
            if self.warmup_method == "linear":
                std_range = min(curr_iter / warmup_iters, 1.0) * (
                    self.std_devs[1] - self.std_devs[0]
                )
            else:
                raise ValueError(f"`warmup_method={self.warmup_method}` not supported")
        else:
            std_range = self.std_devs[1] - self.std_devs[0]

        g = self.generator
        std_dev = self.std_devs[0] + std_range * torch.rand(1, generator=g, device=g.device).item()
        return std_dev

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def forward(self, kspace, mask=None, seed=None, clone=True) -> torch.Tensor:
        """Performs augmentation on undersampled kspace mask."""
        if clone:
            kspace = kspace.clone()
        mask = cplx.get_mask(kspace)

        g = (
            self.generator
            if seed is None
            else torch.Generator(device=kspace.device).manual_seed(seed)
        )
        noise_std = self.choose_std_dev()
        if cplx.is_complex(kspace):
            noise = noise_std * torch.randn(kspace.shape + (2,), generator=g, device=kspace.device)
            noise = torch.view_as_complex(noise)
        else:
            noise = noise_std * torch.randn(kspace.shape, generator=g, device=kspace.device)

        if self.rho is not None and self.rho != 1:
            mask = self.subsample_mask(mask)
        masked_noise = noise * mask
        aug_kspace = kspace + masked_noise

        return aug_kspace

    def subsample_mask(self, mask: torch.Tensor, generator=None):
        """Subsamples mask to add the noise to.

        Currently done uniformly at random.
        """
        rho = self.rho
        shape = mask.shape
        mask = mask.view(-1)

        # TODO: this doesnt work if the matrix is > 2*24 in size.
        # TODO: make this a bit more optimized
        num_valid = torch.sum(mask)
        weights = mask / num_valid
        num_samples = int((1 - rho) * num_valid)
        samples = torch.multinomial(weights, num_samples, replacement=False, generator=generator)
        mask[samples] = 0

        mask = mask.view(shape)
        return mask

    @classmethod
    def from_cfg(cls, cfg, seed=None, **kwargs):
        cfg = cfg.MODEL.CONSISTENCY.AUG.NOISE
        return cls(cfg.STD_DEV, scheduler=cfg.SCHEDULER, mask=cfg.MASK, seed=seed, **kwargs)



from meddlr.metrics.functional import image, sem_seg
from meddlr.metrics.functional.image import l2_norm, mse, nrmse, psnr, rmse, ssim  # noqa: F401
from meddlr.metrics.functional.sem_seg import (  # noqa: F401
    assd,
    average_symmetric_surface_distance,
    coefficient_variation,
    cv,
    dice,
    dice_score,
    voe,
    volumetric_overlap_error,
)

__all__ = []
__all__.extend(image.__all__)
__all__.extend(sem_seg.__all__)



# Copyright 2013-2015. The Regents of the University of California.
# All rights reserved. Use of this source code is governed by
# a BSD-style license which can be found in the LICENSE file.
#
# Authors:
# 2013 Martin Uecker <uecker@eecs.berkeley.edu>
# 2015 Jonathan Tamir <jtamir@eecs.berkeley.edu>

import numpy as np


def read_hdr(name, order="C"):
    """Read hdr file."""
    # get dims from .hdr
    h = open(name + ".hdr", "r")
    h.readline()  # skip line
    l = h.readline()  # noqa: E741
    h.close()
    dims = [int(i) for i in l.split()]
    if order == "C":
        dims.reverse()
    return dims


def read(name, order="C"):
    dims = read_hdr(name, order)

    # remove singleton dimensions from the end
    n = np.prod(dims)
    # dims_prod = np.cumprod(dims)
    # dims = dims[:np.searchsorted(dims_prod, n)+1]

    # load data and reshape into dims
    d = open(name + ".cfl", "r")
    a = np.fromfile(d, dtype=np.complex64, count=n)
    d.close()

    return a.reshape(dims, order=order)


def readcfl(name):
    return read(name, order="F")


def write(name, array, order="C"):
    h = open(name + ".hdr", "w")
    h.write("# Dimensions\n")
    if order == "C":
        for i in array.shape[::-1]:
            h.write("%d " % i)
    else:
        for i in array.shape:
            h.write("%d " % i)
    h.write("\n")
    h.close()

    d = open(name + ".cfl", "w")
    if order == "C":
        array.astype(np.complex64).tofile(d)
    else:
        # tranpose for column-major order
        array.T.astype(np.complex64).tofile(d)
    d.close()


def writecfl(name, array):
    write(name, array, order="F")



"""Training hooks/callbacks.

Adapted from
https://github.com/facebookresearch/detectron2
"""

import datetime
import logging
import os
import tempfile
import time
from collections import Counter

import torch
from fvcore.common.checkpoint import PeriodicCheckpointer as _PeriodicCheckpointer
from fvcore.common.timer import Timer

from meddlr.engine.train_loop import HookBase
from meddlr.evaluation.testing import flatten_results_dict
from meddlr.solver import GradAccumOptimizer
from meddlr.utils import env
from meddlr.utils.events import EventWriter

try:
    from guppy import hpy
except ImportError:  # pragma: no cover
    hpy = None


__all__ = [
    "CallbackHook",
    "IterationTimer",
    "PeriodicWriter",
    "PeriodicCheckpointer",
    "LRScheduler",
    "AutogradProfiler",
    "EvalHook",
    "FlushOptimizer",
]

_PATH_MANAGER = env.get_path_manager()

"""
Implement some common hooks.
"""


class CallbackHook(HookBase):
    """
    Create a hook using callback functions provided by the user.
    """

    def __init__(self, *, before_train=None, after_train=None, before_step=None, after_step=None):
        """
        Each argument is a function that takes one argument: the trainer.
        """
        self._before_train = before_train
        self._before_step = before_step
        self._after_step = after_step
        self._after_train = after_train

    def before_train(self):
        if self._before_train:
            self._before_train(self.trainer)

    def after_train(self):
        if self._after_train:
            self._after_train(self.trainer)
        # The functions may be closures that hold reference to the trainer
        # Therefore, delete them to avoid circular reference.
        del self._before_train, self._after_train
        del self._before_step, self._after_step

    def before_step(self):
        if self._before_step:
            self._before_step(self.trainer)

    def after_step(self):
        if self._after_step:
            self._after_step(self.trainer)


class IterationTimer(HookBase):
    """
    Track the time spent for each iteration (each run_step call in the trainer).
    Print a summary in the end of training.

    This hook uses the time between the call to its :meth:`before_step`
    and :meth:`after_step` methods.
    Under the convention that :meth:`before_step` of all hooks should only
    take negligible amount of time, the :class:`IterationTimer` hook should be
    placed at the beginning of the list of hooks to obtain accurate timing.
    """

    def __init__(self, warmup_iter=3):
        """
        Args:
            warmup_iter (int): the number of iterations at the beginning to
                exclude from timing.
        """
        self._warmup_iter = warmup_iter
        self._step_timer = Timer()

    def before_train(self):
        self._start_time = time.perf_counter()
        self._total_timer = Timer()
        self._total_timer.pause()

    def after_train(self):
        logger = logging.getLogger(__name__)
        total_time = time.perf_counter() - self._start_time
        total_time_minus_hooks = self._total_timer.seconds()
        hook_time = total_time - total_time_minus_hooks

        num_iter = self.trainer.iter + 1 - self.trainer.start_iter - self._warmup_iter

        if num_iter > 0 and total_time_minus_hooks > 0:
            # Speed is meaningful only after warmup
            # NOTE this format is parsed by grep in some scripts
            logger.info(
                "Overall training speed: "
                "{} iterations in {} ({:.4f} s / it)".format(
                    num_iter,
                    str(datetime.timedelta(seconds=int(total_time_minus_hooks))),
                    total_time_minus_hooks / num_iter,
                )
            )

        logger.info(
            "Total training time: {} ({} on hooks)".format(
                str(datetime.timedelta(seconds=int(total_time))),
                str(datetime.timedelta(seconds=int(hook_time))),
            )
        )

    def before_step(self):
        self._step_timer.reset()
        self._total_timer.resume()

    def after_step(self):
        # +1 because we're in after_step
        iter_done = self.trainer.iter - self.trainer.start_iter + 1
        if iter_done >= self._warmup_iter:
            sec = self._step_timer.seconds()
            self.trainer.storage.put_scalars(time=sec)
        else:
            self._start_time = time.perf_counter()
            self._total_timer.reset()

        self._total_timer.pause()


class PeriodicWriter(HookBase):
    """
    Write events to EventStorage periodically.

    It is executed every ``period`` iterations and after the last iteration.
    """

    def __init__(self, writers, periods=(20,)):
        """
        Args:
            writers (list[EventWriter]): a list of EventWriter objects
            periods(`int` or `Sequence[int]`): Period(s) over which to
                save data. This is useful when the evaluation period is
                not a multiple of the default period.
        """
        self._writers = writers
        for w in writers:
            assert isinstance(w, EventWriter), w
        if isinstance(periods, int):
            periods = (periods,)
        self._periods = periods

    def after_step(self):
        if (
            any((self.trainer.iter + 1) % period == 0 for period in self._periods)
            or self.trainer.iter == self.trainer.max_iter - 1
        ):
            for writer in self._writers:
                writer.write()

    def after_train(self):
        for writer in self._writers:
            writer.close()


class PeriodicCheckpointer(_PeriodicCheckpointer, HookBase):
    """
    Same as :class:`meddlr.checkpoint.PeriodicCheckpointer`, but as a hook.

    Note that when used as a hook,
    it is unable to save additional data other than what's defined
    by the given `checkpointer`.

    It is executed every ``period`` iterations and after the last iteration.
    """

    def before_train(self):
        self.max_iter = self.trainer.max_iter

    def after_step(self):
        # No way to use **kwargs
        self.step(self.trainer.iter)


class LRScheduler(HookBase):
    """
    A hook which executes a torch builtin LR scheduler and summarizes the LR.
    It is executed after every iteration.
    """

    def __init__(self, optimizer, scheduler):
        """
        Args:
            optimizer (torch.optim.Optimizer):
            scheduler (torch.optim._LRScheduler)
        """
        self._optimizer = optimizer
        self._scheduler = scheduler

        # NOTE: some heuristics on what LR to summarize
        # summarize the param group with most parameters
        largest_group = max(len(g["params"]) for g in optimizer.param_groups)

        if largest_group == 1:
            # If all groups have one parameter,
            # then find the most common initial LR, and use it for summary
            lr_count = Counter([g["lr"] for g in optimizer.param_groups])
            lr = lr_count.most_common()[0][0]
            for i, g in enumerate(optimizer.param_groups):
                if g["lr"] == lr:
                    self._best_param_group_id = i
                    break
        else:
            for i, g in enumerate(optimizer.param_groups):
                if len(g["params"]) == largest_group:
                    self._best_param_group_id = i
                    break

    def after_step(self):
        lr = self._optimizer.param_groups[self._best_param_group_id]["lr"]
        self.trainer.storage.put_scalar("lr", lr, smoothing_hint=False)
        self._scheduler.step()


class AutogradProfiler(HookBase):
    """
    A hook which runs `torch.autograd.profiler.profile`.

    Examples:

    .. code-block:: python

        hooks.AutogradProfiler(
             lambda trainer: trainer.iter > 10 and trainer.iter < 20,
                self.cfg.OUTPUT_DIR
        )

    The above example will run the profiler for iteration 10~20 and dump
    results to ``OUTPUT_DIR``. We did not profile the first few iterations
    because they are typically slower than the rest.
    The result files can be loaded in the ``chrome://tracing`` page in chrome
    browser.

    Note:
        When used together with NCCL on older version of GPUs,
        autograd profiler may cause deadlock because it unnecessarily allocates
        memory on every device it sees. The memory management calls, if
        interleaved with NCCL calls, lead to deadlock on GPUs that do not
        support `cudaLaunchCooperativeKernelMultiDevice`.
    """

    def __init__(self, enable_predicate, output_dir, *, use_cuda=True):
        """
        Args:
            enable_predicate (callable[trainer -> bool]): a function which takes
                a trainer, and returns whether to enable the profiler.
                It will be called once every step, and can be used to select
                which steps to profile.
            output_dir (str): the output directory to dump tracing files.
            use_cuda (bool): same as in `torch.autograd.profiler.profile`.
        """
        self._enable_predicate = enable_predicate
        self._use_cuda = use_cuda
        self._output_dir = output_dir

    def before_step(self):
        if self._enable_predicate(self.trainer):
            self._profiler = torch.autograd.profiler.profile(use_cuda=self._use_cuda)
            self._profiler.__enter__()
        else:
            self._profiler = None

    def after_step(self):
        if self._profiler is None:
            return
        self._profiler.__exit__(None, None, None)
        out_file = os.path.join(
            self._output_dir, "profiler-trace-iter{}.json".format(self.trainer.iter)
        )
        if "://" not in out_file:
            self._profiler.export_chrome_trace(out_file)
        else:
            # Support non-posix filesystems
            with tempfile.TemporaryDirectory(prefix="detectron2_profiler") as d:
                tmp_file = os.path.join(d, "tmp.json")
                self._profiler.export_chrome_trace(tmp_file)
                with open(tmp_file) as f:
                    content = f.read()
            with _PATH_MANAGER.open(out_file, "w") as f:
                f.write(content)


class FlushOptimizer(HookBase):
    """Flush optimizer after period(s). Typically done before eval."""

    def __init__(self, optimizer: GradAccumOptimizer, period):
        self._period = period
        self._optimizer = optimizer

    def after_step(self):
        next_iter = self.trainer.iter + 1
        is_final = next_iter == self.trainer.max_iter
        if is_final or (self._period > 0 and next_iter % self._period == 0):
            self._optimizer.flush()


class EvalHook(HookBase):
    """
    Run an evaluation function periodically, and at the end of training.

    It is executed every ``eval_period`` iterations and after the last
    iteration.
    """

    def __init__(self, eval_period, eval_function):
        """
        Args:
            eval_period (int): the period to run `eval_function`.
            eval_function (callable): a function which takes no arguments, and
                returns a nested dict of evaluation metrics.

        Note:
            This hook must be enabled in all or none workers.
            If you would like only certain workers to perform evaluation,
            give other workers a no-op function (`eval_function=lambda: None`).
        """
        self._period = eval_period
        self._func = eval_function
        self._done_eval_at_last = False

    def _do_eval(self):
        start_time = time.perf_counter()

        results = self._func()

        if results:
            assert isinstance(
                results, dict
            ), "Eval function must return a dict. Got {} instead.".format(results)

            flattened_results = flatten_results_dict(results)
            for k, v in flattened_results.items():
                try:
                    v = float(v)
                except Exception:
                    raise ValueError(
                        "[EvalHook] eval_function should return "
                        "a nested dict of float. "
                        "Got '{}: {}' instead.".format(k, v)
                    )
            self.trainer.storage.put_scalars(**flattened_results, smoothing_hint=False)

        time_elapsed = time.perf_counter() - start_time  # in seconds
        self.trainer.storage.put_scalars(eval_time=time_elapsed, smoothing_hint=False)

    def after_step(self):
        next_iter = self.trainer.iter + 1
        is_final = next_iter == self.trainer.max_iter
        if is_final or (self._period > 0 and next_iter % self._period == 0):
            self._do_eval()
            if is_final:
                self._done_eval_at_last = True

    def after_train(self):
        if not self._done_eval_at_last:
            self._do_eval()
        # func is likely a closure that holds reference to the trainer
        # therefore we clean it to avoid circular reference in the end
        del self._func


class MemoryProfiler(HookBase):
    """
    Profile RAM.

    This class relies on the ``guppy`` module (pip install guppy3).

    Note:
        This function is still in ALPHA.
    """

    def __init__(self, period):
        if hpy is None:
            raise ModuleNotFoundError("Module `guppy` not found. Install with `pip install guppy3`")
        self._period = period

    def after_step(self):
        next_iter = self.trainer.iter + 1
        is_final = next_iter == self.trainer.max_iter
        if is_final or (self._period > 0 and next_iter % self._period == 0):
            hp = hpy()
            heap = hp.heap()
            self.trainer.storage.put_scalar(
                "guppy/total_RAM", heap.size / 1024.0 / 1024.0, smoothing_hint=False
            )



"""
Copyright (c) Facebook, Inc. and its affiliates.
This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

import torch
import torchvision.utils as tv_utils
from torch import nn
from torch.nn import functional as F

import meddlr.ops.complex as cplx
from meddlr.config.config import configurable
from meddlr.utils import transforms as T
from meddlr.utils.events import get_event_storage
from meddlr.utils.transforms import SenseModel

from .build import META_ARCH_REGISTRY

__all__ = ["UnetModel"]


class ConvBlock(nn.Module):
    """
    A Convolutional Block that consists of two convolution layers each followed by
    instance normalization, LeakyReLU activation and dropout.
    """

    def __init__(self, in_chans, out_chans, drop_prob):
        """
        Args:
            in_chans (int): Number of channels in the input.
            out_chans (int): Number of channels in the output.
            drop_prob (float): Dropout probability.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.drop_prob = drop_prob

        self.layers = nn.Sequential(
            nn.Conv2d(in_chans, out_chans, kernel_size=3, padding=1, bias=False),
            nn.InstanceNorm2d(out_chans),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
            nn.Dropout2d(drop_prob),
            nn.Conv2d(out_chans, out_chans, kernel_size=3, padding=1, bias=False),
            nn.InstanceNorm2d(out_chans),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
            nn.Dropout2d(drop_prob),
        )

    def forward(self, input):
        """
        Args:
            input (torch.Tensor): Input tensor of shape [batch_size, self.in_chans, height, width]
        Returns:
            (torch.Tensor): Output tensor of shape [batch_size, self.out_chans, height, width]
        """
        return self.layers(input)

    def __repr__(self):
        return (
            f"ConvBlock(in_chans={self.in_chans}, out_chans={self.out_chans}, "
            f"drop_prob={self.drop_prob})"
        )


class TransposeConvBlock(nn.Module):
    """
    A Transpose Convolutional Block that consists of one convolution transpose layers followed by
    instance normalization and LeakyReLU activation.
    """

    def __init__(self, in_chans, out_chans):
        """
        Args:
            in_chans (int): Number of channels in the input.
            out_chans (int): Number of channels in the output.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans

        self.layers = nn.Sequential(
            nn.ConvTranspose2d(in_chans, out_chans, kernel_size=2, stride=2, bias=False),
            nn.InstanceNorm2d(out_chans),
            nn.LeakyReLU(negative_slope=0.2, inplace=True),
        )

    def forward(self, input):
        """
        Args:
            input (torch.Tensor): Input tensor of shape [batch_size, self.in_chans, height, width]
        Returns:
            (torch.Tensor): Output tensor of shape [batch_size, self.out_chans, height, width]
        """
        return self.layers(input)

    def __repr__(self):
        return f"ConvBlock(in_chans={self.in_chans}, out_chans={self.out_chans})"


@META_ARCH_REGISTRY.register()
class UnetModel(nn.Module):
    """
    PyTorch implementation of a U-Net model.
    This is based on:
        Olaf Ronneberger, Philipp Fischer, and Thomas Brox. U-net: Convolutional networks
        for biomedical image segmentation. In International Conference on Medical image
        computing and computer-assisted intervention, pages 234–241. Springer, 2015.
    """

    @configurable
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        channels: int = 32,
        num_pool_layers: int = 4,
        dropout: float = 0.0,
        use_latent: bool = False,
        num_latent_layers: int = 1,
        normalize: bool = False,
        vis_period: int = -1,
    ):
        """
        Args:
            in_chans (int): Number of channels in the input to the U-Net model.
            out_chans (int): Number of channels in the output to the U-Net model.
            chans (int): Number of output channels of the first convolution layer.
            num_pool_layers (int): Number of down-sampling and up-sampling layers.
            drop_prob (float): Dropout probability.
            normalize (bool, optional): If ``True``, normalize the input for stability.
        """
        super().__init__()

        self.in_chans = in_channels
        self.out_chans = out_channels
        self.chans = channels
        self.num_pool_layers = num_pool_layers
        self.drop_prob = dropout
        self.use_latent = use_latent
        self.num_latent_layers = num_latent_layers
        self.normalize = normalize

        self.down_sample_layers = nn.ModuleList([ConvBlock(in_channels, channels, dropout)])
        ch = channels
        for _i in range(num_pool_layers - 1):
            self.down_sample_layers += [ConvBlock(ch, ch * 2, dropout)]
            ch *= 2
        self.conv = ConvBlock(ch, ch * 2, dropout)

        self.up_conv = nn.ModuleList()
        self.up_transpose_conv = nn.ModuleList()
        for _i in range(num_pool_layers - 1):
            self.up_transpose_conv += [TransposeConvBlock(ch * 2, ch)]
            self.up_conv += [ConvBlock(ch * 2, ch, dropout)]
            ch //= 2

        self.up_transpose_conv += [TransposeConvBlock(ch * 2, ch)]
        self.up_conv += [
            nn.Sequential(
                ConvBlock(ch * 2, ch, dropout),
                nn.Conv2d(ch, self.out_chans, kernel_size=1, stride=1),
            )
        ]

        self.vis_period = vis_period

    def register_hooks(self):
        if self.use_latent:
            self.feats = {}
            self.hooks = []
            self.hooks.append(self.conv.register_forward_hook(self.get_latent("E4")))
            for _i in range(self.num_latent_layers - 1):
                k = self.num_pool_layers - 1 - _i
                self.hooks.append(
                    self.up_conv[_i].register_forward_hook(self.get_latent("D" + str(k)))
                )
                self.hooks.append(
                    self.down_sample_layers[k].register_forward_hook(self.get_latent("E" + str(k)))
                )

    def remove_hooks(self):
        for _i in range(self.num_latent_layers - 1):
            self.hooks[_i].remove()

    def get_latent(self, layer_name):
        def hook(module, input, output):
            self.feats[layer_name] = output

        return hook

    def visualize_training(self, kspace, zfs, targets, preds):
        """A function used to visualize reconstructions.

        TODO: Refactor out

        Args:
            targets: NxHxWx2 tensors of target images.
            preds: NxHxWx2 tensors of predictions.
        """
        storage = get_event_storage()

        with torch.no_grad():
            if cplx.is_complex(kspace):
                kspace = torch.view_as_real(kspace)
            if cplx.is_complex(targets) and not cplx.is_complex(zfs):
                # Zero-filled needs to be manually converted.
                zfs = torch.view_as_complex(zfs)
            kspace = kspace[0, ..., 0, :].unsqueeze(0).cpu()  # calc mask for first coil only
            targets = targets[0, ...].unsqueeze(0).cpu()
            preds = preds[0, ...].unsqueeze(0).cpu()
            zfs = zfs[0, ...].unsqueeze(0).cpu()

            all_images = torch.cat([zfs, preds, targets], dim=2)

            imgs_to_write = {
                "phases": cplx.angle(all_images),
                "images": cplx.abs(all_images),
                "errors": cplx.abs(preds - targets),
                "masks": cplx.get_mask(kspace),
            }

            for name, data in imgs_to_write.items():
                data = data.squeeze(-1).unsqueeze(1)
                data = tv_utils.make_grid(data, nrow=1, padding=1, normalize=True, scale_each=True)
                storage.put_image("train/{}".format(name), data.numpy(), data_format="CHW")

    def norm(self, x):
        """
        Args:
            x (torch.Tensor): Input tensor. Shape [B, 2*C, H, W, ...]
        """
        ndim = x.ndim
        x = x.view(x.shape[:2] + (-1,))

        mean = x.mean(dim=-1).reshape(x.shape[:2] + (1,) * (ndim - 2))
        std = x.std(dim=-1).reshape(x.shape[:2] + (1,) * (ndim - 2))

        return (x - mean) / std, mean, std

    def unnorm(self, x, mean, std):
        return x * std + mean

    def base_forward(self, zf_image):
        """The base forward function.

        Args:
            zf_image (torch.Tensor): Zero-filled image. Shape [B, C, ..., H, W].
                The real and imaginary channels should be part of the channels
                dimension if the input is complex and the model does not have
                complex weights.

        Returns:
            torch.Tensor: Reconstructed image. Shape [B, C, ..., H, W].
        """
        x = zf_image

        stack = []
        if self.normalize:
            x, mean, std = self.norm(x)

        # Apply down-sampling layers
        for layer in self.down_sample_layers:
            x = layer(x)
            stack.append(x)
            x = F.avg_pool2d(x, kernel_size=2, stride=2, padding=0)

        x = self.conv(x)

        # Apply up-sampling layers
        for transpose_conv, conv in zip(self.up_transpose_conv, self.up_conv):
            downsample_layer = stack.pop()
            x = transpose_conv(x)

            # Reflect pad on the right/botton if needed to handle odd input dimensions.
            padding = [0, 0, 0, 0]
            if x.shape[-1] != downsample_layer.shape[-1]:
                padding[1] = 1  # Padding right
            if x.shape[-2] != downsample_layer.shape[-2]:
                padding[3] = 1  # Padding bottom
            if sum(padding) != 0:
                x = F.pad(x, padding, "reflect")

            x = torch.cat([x, downsample_layer], dim=1)
            x = conv(x)

        if self.normalize:
            x = self.unnorm(x, mean, std)

        return x

    def forward(self, input, return_pp=False, vis_training=False):
        """
        Args:
            input (torch.Tensor): Input tensor of shape [batch_size, self.in_chans, height, width]
        Returns:
            (torch.Tensor): Output tensor of shape [batch_size, self.out_chans, height, width]
        """
        self.register_hooks()

        # Need to fetch device at runtime for proper data transfer.
        inputs = input
        device = next(self.parameters()).device
        kspace = inputs["kspace"].to(device)
        target = inputs["target"].to(device) if "target" in inputs else None
        mask = inputs["mask"].to(device) if "mask" in inputs else None
        A = inputs["signal_model"].to(device) if "signal_model" in inputs else None
        maps = inputs["maps"].to(device)

        if mask is None:
            mask = cplx.get_mask(kspace)
        kspace *= mask

        # Declare signal model.
        if A is None:
            A = SenseModel(maps, weights=mask)

        # Zero-filled Sense Recon.
        if cplx.is_complex(maps):
            zf_image = A(kspace, adjoint=True)
        # Zero-filled RSS Recon.
        else:
            zf_image_init = T.ifft2(kspace)
            zf_image_rss = torch.sqrt(torch.sum(cplx.abs(zf_image_init) ** 2, axis=-1))
            zf_image = torch.complex(zf_image_rss, torch.zeros_like(zf_image_rss)).unsqueeze(-1)

        use_cplx = cplx.is_complex(zf_image)
        if use_cplx:
            zf_image = torch.view_as_real(zf_image)

        num_maps = zf_image.shape[-2]
        if num_maps > 1:
            output = zf_image.reshape(zf_image.shape[:-2] + (-1,)).unsqueeze(-2)
        else:
            output = zf_image
        output = zf_image.permute(0, 4, 1, 2, 3).squeeze(-1)

        # Run U-Net.
        output = self.base_forward(output)

        if num_maps > 1:
            pred = output.permute(0, 2, 3, 1)
            pred = pred.reshape(pred.shape[:-1] + (num_maps, 2))
        else:
            pred = output.unsqueeze(-1).permute(0, 2, 3, 4, 1)

        if use_cplx:
            pred = torch.view_as_complex(pred.contiguous())

        output_dict = {"pred": pred, "target": target}  # N x Y x Z x 1 (x 2)  # N x Y x Z x 1 (x 2)

        if return_pp:
            output_dict.update({k: inputs[k] for k in ["mean", "std", "norm"]})

        # TODO (arjundd): Figure out why this is needed during SSDU training.
        images_available = all(x is not None for x in [kspace, zf_image, target, pred])
        if images_available and self.training and (vis_training or self.vis_period > 0):
            storage = get_event_storage()
            if vis_training or storage.iter % self.vis_period == 0:
                self.visualize_training(kspace, zf_image, target, pred)

        if use_cplx:
            zf_image = torch.view_as_complex(zf_image)
        output_dict["zf_image"] = zf_image

        if self.use_latent:
            output_dict["latent"] = self.feats

        self.remove_hooks()

        return output_dict

    @classmethod
    def from_config(cls, cfg):
        return {
            "in_channels": cfg.MODEL.UNET.IN_CHANNELS,
            "out_channels": cfg.MODEL.UNET.OUT_CHANNELS,
            "channels": cfg.MODEL.UNET.CHANNELS,
            "num_pool_layers": cfg.MODEL.UNET.NUM_POOL_LAYERS,
            "dropout": cfg.MODEL.UNET.DROPOUT,
            "use_latent": cfg.get_recursive("MODEL.CONSISTENCY.USE_LATENT", False),
            "num_latent_layers": cfg.get_recursive("MODEL.CONSISTENCY.NUM_LATENT_LAYERS", 1),
            "normalize": cfg.MODEL.UNET.NORMALIZE,
            "vis_period": cfg.VIS_PERIOD,
        }



from typing import Sequence, Tuple, Union

import torch

import meddlr.transforms.functional as stf
from meddlr.transforms.build import TRANSFORM_REGISTRY
from meddlr.transforms.transform import Transform


@TRANSFORM_REGISTRY.register()
class MRIMotionTransform(Transform):
    """A model that corrupts kspace inputs with motion.
    Motion is a common artifact experienced during the MR imaging forward problem.
    When a patient moves, the recorded (expected) location of the kspace sample is
    different than the actual location where the kspace sample that was acquired.
    This module is responsible for simulating different motion artifacts.
    Args:
        seed (int, optional): The fixed seed.
    Attributes:
        generator (torch.Generator): The generator that should be used for all
            random logic in this class.
    Things to consider:
        1. What other information is relevant for inducing motion corruption?
            This could include:
            - ``traj``: The scan trajectory
            - ``etl``: The echo train length - how many readouts per shot.
            - ``num_shots``: Number of shots.
        2. What would a simple translational motion model look?
    Note:
        We do not store this as a module or else it would be saved to the model
        definition, which we dont want.
    """

    def __init__(
        self,
        std_dev: Union[float, Sequence[float]],
        seed: int = None,
        generator: torch.Generator = None,
    ):
        self.std_dev = std_dev
        self.seed = seed

        gen_state = None
        if generator is not None:
            gen_state = generator.get_state()
        self._generator_state = gen_state

    def _generator(self, data: torch.Tensor):
        seed = self.seed

        g = torch.Generator(device=data.device)
        if seed is None:
            g.set_state(self._generator_state)
        else:
            g = g.manual_seed(seed)
        return g

    def apply_kspace(self, kspace, channel_first: bool = True) -> torch.Tensor:
        """Performs motion corruption on kspace image.

        Args:
            kspace (torch.Tensor): The complex tensor. Shape ``(N, #coils, Y, X, [2])``.

        Returns:
            torch.Tensor: The motion corrupted kspace.

        Note:
            For backwards compatibility with torch<1.6, complex tensors may also have the shape
            ``(..., 2)``, where the 2 channels in the last dimension are real and
            imaginary, respectively.
            TODO: This code should account for that case as well.
        """
        scale = self.std_dev
        g = self._generator(kspace)
        return stf.add_even_odd_motion(
            kspace, scale=scale, channel_first=channel_first, generator=g
        )

    def _eq_attrs(self) -> Tuple[str]:
        return ("std_dev", "seed", "_generator_state")



"""CfgNode implementation.
"""
import functools
import inspect
import logging
import re
from typing import Any, Mapping

import numpy as np
from fvcore.common.config import CfgNode as _CfgNode


class CfgNode(_CfgNode):
    """
    The same as `fvcore.common.config.CfgNode`, but different in:

    1. Use unsafe yaml loading by default.
      Note that this may lead to arbitrary code execution: you must not
      load a config file from untrusted sources before manually inspecting
      the content of the file.
    2. Support config versioning.
      When attempting to merge an old config, it will convert the old config
      automatically.
    """

    # Note that the default value of allow_unsafe is changed to True
    def merge_from_file(self, cfg_filename: str, allow_unsafe: bool = True) -> None:
        """
        Adapted from https://github.com/facebookresearch/detectron2
        """
        loaded_cfg = _CfgNode.load_yaml_with_base(cfg_filename, allow_unsafe=allow_unsafe)
        loaded_cfg = type(self)(loaded_cfg)

        # defaults.py needs to import CfgNode
        from .defaults import _C

        latest_ver = _C.VERSION
        assert latest_ver == self.VERSION, (
            "CfgNode.merge_from_file is only allowed on a config " "object of latest version!"
        )

        logger = logging.getLogger(__name__)

        loaded_ver = loaded_cfg.get("VERSION", None)
        if loaded_ver is None:
            from .compat import guess_version

            loaded_ver = guess_version(loaded_cfg, cfg_filename)
        assert loaded_ver <= self.VERSION, "Cannot merge a v{} config into a v{} config.".format(
            loaded_ver, self.VERSION
        )

        if loaded_ver == self.VERSION:
            self.merge_from_other_cfg(loaded_cfg)
        else:
            # compat.py needs to import CfgNode
            from .compat import downgrade_config, upgrade_config

            logger.warning(
                "Loading an old v{} config file '{}' by "
                "automatically upgrading to v{}. "
                "See docs/CHANGELOG.md for instructions to "
                "update your files.".format(loaded_ver, cfg_filename, self.VERSION)
            )
            # To convert, first obtain a full config at an old version
            old_self = downgrade_config(self, to_version=loaded_ver)
            old_self.merge_from_other_cfg(loaded_cfg)
            new_config = upgrade_config(old_self)
            self.clear()
            self.update(new_config)

    def format_fields(self, unroll: bool = False):
        """
        Format string fields in the config by filling them in
        with different parameter values.

        The operation is done in place.

        Args:
            unroll (bool, optional): If ``True``, sequence types (e.g. tuple, list)
                stringified as ``seq[0]-seq[1]-...``. Dict types will be
                stringified as ``k[0]=v[0]-k[1]=v[1]-...``.

        Returns:
            CfgNode: self
        """
        return format_config_fields(self, unroll=unroll, inplace=True)

    def get_recursive(self, key, default: Any = np._NoValue):
        """Get a key recursively from the config.

        Args:
            key (str): The dot-separated key.
            default (Any, optional): The value to return if the key is not found.
                If not provided, a KeyError will be raised.

        Returns:
            object: The value corresponding to the key.

        Raises:
            KeyError: If the key is not found and no default was provided.
        """
        d = self
        try:
            for k in key.split("."):
                d = d[k]
        except KeyError as e:
            if default != np._NoValue:
                return default
            raise e
        return d

    def set_recursive(self, name: str, value: Any):
        """Set a key recursively in the config.

        Args:
            name (str): The dot-separated key.
            value (object): The value to set.
        """
        cfg = self
        keys = name.split(".")
        for k in keys[:-1]:
            cfg = cfg[k]

        setattr(cfg, keys[-1], value)

    def dump(self, *args, **kwargs):
        """
        Returns:
            str: a yaml string representation of the config
        """
        # to make it show up in docs
        return super().dump(*args, **kwargs)

    def freeze(self):
        """Make this CfgNode and all of its children immutable.

        Returns:
            CfgNode: self
        """
        super().freeze()
        return self

    def defrost(self):
        """Make this CfgNode and all of its children mutable.

        Returns:
            CfgNode: self
        """
        super().defrost()
        return self


global_cfg = CfgNode()
_base_cfg = None


def get_cfg() -> CfgNode:
    """
    Get a copy of the default config.

    Returns:
        a meddlr CfgNode instance.
    """
    from .defaults import _C

    cfg = _C if _base_cfg is None else _base_cfg
    return cfg.clone()


def set_cfg(cfg: CfgNode) -> None:
    """Set the base config object to use.

    This is useful when customizing ss_recon for different projects.

    Args:
        cfg (CfgNode): Set the base config.
    """
    global _base_cfg
    _base_cfg = cfg


def set_global_cfg(cfg: CfgNode) -> None:
    """
    Let the global config point to the given cfg.

    Assume that the given "cfg" has the key "KEY", after calling
    `set_global_cfg(cfg)`, the key can be accessed by:

    .. code-block:: python

        from meddlr.config import global_cfg
        print(global_cfg.KEY)

    By using a hacky global config, you can access these configs anywhere,
    without having to pass the config object or the values deep into the code.
    This is a hacky feature introduced for quick prototyping / research
    exploration.
    """
    global global_cfg
    global_cfg.clear()
    global_cfg.update(cfg)


def configurable(init_func=None, *, from_config=None):
    """
    Decorate a function or a class's __init__ method so that it can be called
    with a :class:`CfgNode` object using a :func:`from_config` function that translates
    :class:`CfgNode` to arguments.

    Adapted from https://github.com/facebookresearch/detectron2

    Examples:
    ::
        # Usage 1: Decorator on __init__:
        class A:
            @configurable
            def __init__(self, a, b=2, c=3):
                pass
            @classmethod
            def from_config(cls, cfg):   # 'cfg' must be the first argument
                # Returns kwargs to be passed to __init__
                return {"a": cfg.A, "b": cfg.B}
        a1 = A(a=1, b=2)  # regular construction
        a2 = A(cfg)       # construct with a cfg
        a3 = A(cfg, b=3, c=4)  # construct with extra overwrite
        # Usage 2: Decorator on any function. Needs an extra from_config argument:
        @configurable(from_config=lambda cfg: {"a: cfg.A, "b": cfg.B})
        def a_func(a, b=2, c=3):
            pass
        a1 = a_func(a=1, b=2)  # regular call
        a2 = a_func(cfg)       # call with a cfg
        a3 = a_func(cfg, b=3, c=4)  # call with extra overwrite
    Args:
        init_func (callable): a class's ``__init__`` method in usage 1. The
            class must have a ``from_config`` classmethod which takes `cfg` as
            the first argument.
        from_config (callable): the from_config function in usage 2. It must take `cfg`
            as its first argument.
    """

    if init_func is not None:
        assert (
            inspect.isfunction(init_func)
            and from_config is None
            and init_func.__name__ == "__init__"
        ), "Incorrect use of @configurable. Check API documentation for examples."

        @functools.wraps(init_func)
        def wrapped(self, *args, **kwargs):
            try:
                from_config_func = type(self).from_config
            except AttributeError as e:
                raise AttributeError(
                    "Class with @configurable must have a 'from_config' classmethod."
                ) from e
            if not inspect.ismethod(from_config_func):
                raise TypeError("Class with @configurable must have a 'from_config' classmethod.")

            if _called_with_cfg(*args, **kwargs):
                explicit_args = _get_args_from_config(from_config_func, *args, **kwargs)
                init_func(self, **explicit_args)
            else:
                init_func(self, *args, **kwargs)

        return wrapped

    else:
        if from_config is None:
            return configurable  # @configurable() is made equivalent to @configurable
        assert inspect.isfunction(
            from_config
        ), "from_config argument of configurable must be a function!"

        def wrapper(orig_func):
            @functools.wraps(orig_func)
            def wrapped(*args, **kwargs):
                if _called_with_cfg(*args, **kwargs):
                    explicit_args = _get_args_from_config(from_config, *args, **kwargs)
                    return orig_func(**explicit_args)
                else:
                    return orig_func(*args, **kwargs)

            return wrapped

        return wrapper


def _get_args_from_config(from_config_func, *args, **kwargs):
    """
    Use `from_config` to obtain explicit arguments.
    Returns:
        dict: arguments to be used for cls.__init__
    """
    signature = inspect.signature(from_config_func)
    if list(signature.parameters.keys())[0] != "cfg":
        if inspect.isfunction(from_config_func):
            name = from_config_func.__name__
        else:
            name = f"{from_config_func.__self__}.from_config"
        raise TypeError(f"{name} must take 'cfg' as the first argument!")
    support_var_arg = any(
        param.kind in [param.VAR_POSITIONAL, param.VAR_KEYWORD]
        for param in signature.parameters.values()
    )
    if support_var_arg:  # forward all arguments to from_config, if from_config accepts them
        ret = from_config_func(*args, **kwargs)
    else:
        # forward supported arguments to from_config
        supported_arg_names = set(signature.parameters.keys())
        extra_kwargs = {}
        for name in list(kwargs.keys()):
            if name not in supported_arg_names:
                extra_kwargs[name] = kwargs.pop(name)
        ret = from_config_func(*args, **kwargs)
        # forward the other arguments to __init__
        ret.update(extra_kwargs)
    return ret


def _called_with_cfg(*args, **kwargs):
    """
    Returns:
        bool: whether the arguments contain CfgNode and should be considered
            forwarded to from_config.
    """
    from omegaconf import DictConfig

    if len(args) and isinstance(args[0], (_CfgNode, DictConfig)):
        return True
    if isinstance(kwargs.pop("cfg", None), (_CfgNode, DictConfig)):
        return True
    # `from_config`'s first argument is forced to be "cfg".
    # So the above check covers all cases.
    return False


def _find_format_str_keys(cfg: Mapping, prefix="", accum=()):
    accum = set(accum)
    for k, v in cfg.items():
        k_prefix = prefix + "." + k if prefix else k
        if isinstance(v, Mapping):
            accum |= _find_format_str_keys(cfg[k], prefix=k_prefix, accum=accum)
        elif isinstance(v, str) and (
            (v.startswith('f"') and v.endswith('"')) or (v.startswith("f'") and v.endswith("'"))
        ):
            accum |= {(k_prefix, v)}
    return accum


def _unroll_value_to_str(value) -> str:
    if isinstance(value, (tuple, list)):
        return "-".join(_unroll_value_to_str(v) for v in value)
    elif isinstance(value, dict):
        return "-".join(f"{k}={_unroll_value_to_str(v)}" for k, v in value.items())
    else:
        return str(value)


def format_config_fields(cfg: CfgNode, unroll=False, inplace=False):
    keys_and_val_str = _find_format_str_keys(cfg)
    values_list = []
    for k, val_str in keys_and_val_str:
        start = [x.start() for x in re.finditer("\{", val_str)]
        end = [x.start() for x in re.finditer("\}", val_str)]
        assert len(start) == len(end), f"Could not determine formatting string: {val_str}"
        cfg_keys_to_search = [val_str[s + 1 : e] for s, e in zip(start, end)]
        values = [cfg.get_recursive(v) for v in cfg_keys_to_search]
        if unroll:
            values = [_unroll_value_to_str(v) for v in values]

        fmt_str = ""
        idxs = [0] + [y for x in zip(start, end) for y in x] + [len(val_str)]
        for i in range(len(idxs) // 2):
            fmt_str += val_str[idxs[2 * i] : idxs[2 * i + 1] + 1]
        fmt_str = eval(fmt_str.format(*values))
        values_list.extend([k, fmt_str])
    if not inplace:
        cfg.clone()
    cfg.defrost().merge_from_list(values_list)
    return cfg



from typing import Sequence, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from meddlr.modeling.layers.build import CUSTOM_LAYERS_REGISTRY

__all__ = ["gaussian", "get_gaussian_kernel", "GaussianBlur"]


def gaussian(window_size, sigma, normalize=True):
    def gauss_fcn(x):
        center = window_size // 2 if window_size % 2 == 1 else window_size // 2 - 0.5
        return -((x - center) ** 2) / float(2 * sigma ** 2)

    gauss = torch.stack([torch.exp(torch.tensor(gauss_fcn(x))) for x in range(window_size)])
    if normalize:
        gauss = gauss / gauss.sum()
    return gauss


def get_gaussian_kernel(
    kernel_size: Union[int, Sequence[int]],
    sigma: Union[float, Sequence[float]],
    normalize: bool = True,
) -> torch.Tensor:
    """Function that returns Gaussian filter coefficients.

    Args:
        kernel_size (int(s)): filter size. It should be positive.
        sigma (float(s)): gaussian standard deviation.
        normalize (bool, optional): If `True`, kernel will be normalized. i.e. `kernel.sum() == 1)

    Returns:
        Tensor: nD tensor with gaussian filter coefficients. Shape :math:`(\text{kernel_size})`

    Examples::

        >>> medsegpy.layers.get_gaussian_kernel(3, 2.5)
        tensor([0.3243, 0.3513, 0.3243])

        >>> medsegpy.layers.get_gaussian_kernel((3, 3), (1.5, 1.5))
        tensor([[0.0947, 0.1183, 0.0947],
                [0.1183, 0.1478, 0.1183],
                [0.0947, 0.1183, 0.0947]])
    """
    kernel_size_seq = (kernel_size,) if not isinstance(kernel_size, Sequence) else kernel_size
    sigma_seq = (sigma,) if not isinstance(sigma, Sequence) else sigma
    if not isinstance(kernel_size, (int, Tuple)) or any(k <= 0 for k in kernel_size_seq):
        raise TypeError(
            "kernel_size must be a (sequence of) odd positive integer. "
            "Got {}".format(kernel_size)
        )
    if len(kernel_size_seq) != len(sigma_seq):
        raise ValueError(
            "kernel_size and sigma must have same number of elements. "
            "Got kernel_size={}, sigma={}".format(kernel_size, sigma)
        )
    assert len(kernel_size_seq) <= 26
    kernels_1d = tuple(
        gaussian(ksize, sigma, normalize) for ksize, sigma in zip(kernel_size_seq, sigma_seq)
    )
    elems = tuple(chr(ord("a") + i) for i in range(len(kernels_1d)))
    equation = "{}->{}".format(",".join(elems), "".join(elems))
    return torch.einsum(equation, *kernels_1d)


@CUSTOM_LAYERS_REGISTRY.register()
class GaussianBlur(nn.Module):
    r"""Creates an operator that blurs a tensor using a Gaussian filter.

    The operator smooths the given tensor with a gaussian kernel by convolving
    it to each channel. It suports batched operation.

    Arguments:
        kernel_size (Tuple[int, int]): the size of the kernel. Dimensions are `([D, H], W)`
        sigma (Tuple[float, float]): the standard deviation of the kernel.
            Dimensions are `([D, H], W)`

    Returns:
        Tensor: the blurred tensor. Shape :math:`(B, C, [..., H], W)`

    Shape:
        - Input: :math:`(B, C, H, W)`
        - Output: :math:`(B, C, H, W)`

    Examples:

        >>> input = torch.rand(2, 4, 5, 5)
        >>> gauss = GaussianBlur((3, 3), (1.5, 1.5))
        >>> output = gauss(input)  # 2x4x5x5
    """

    def __init__(
        self, kernel_size: Union[int, Tuple[int, ...]], sigma: Union[float, Tuple[float, ...]]
    ):
        super().__init__()
        kernel_size = (
            (kernel_size,) if not isinstance(kernel_size, Sequence) else tuple(kernel_size)
        )
        if any(k % 2 == 0 for k in kernel_size):
            raise ValueError("kernel_size must be odd and positive. Got {}".format(kernel_size))
        sigma = (sigma,) if not isinstance(sigma, Sequence) else tuple(sigma)
        self.kernel_size: Tuple[int, ...] = kernel_size
        self.sigma: Tuple[float, ...] = sigma
        self._padding: Tuple[int, ...] = self.compute_zero_padding(kernel_size)
        self.kernel: torch.Tensor = nn.Parameter(
            get_gaussian_kernel(kernel_size, sigma, normalize=True),
            requires_grad=False,
        )
        self.spatial_dim = len(kernel_size)
        self.conv = [F.conv1d, F.conv2d, F.conv3d][self.spatial_dim - 1]

    @staticmethod
    def compute_zero_padding(kernel_size: Tuple[int, ...]) -> Tuple[int, ...]:
        """Computes zero padding tuple."""
        computed = [(k - 1) // 2 for k in kernel_size]
        return tuple(computed)

    def forward(self, x: torch.Tensor):
        if not torch.is_tensor(x):
            raise TypeError("Input x type is not a torch.Tensor. Got {}".format(type(x)))

        c = x.shape[1]
        spatial_dim = self.spatial_dim
        tmp_kernel: torch.Tensor = self.kernel
        kernel: torch.Tensor = tmp_kernel.repeat([c, 1] + [1] * spatial_dim)

        # TODO: explore solution when using jit.trace since it raises a warning
        # because the shape is converted to a tensor instead to a int.
        # convolve tensor with gaussian kernel
        return self.conv(x, kernel, padding=self._padding, stride=1, groups=c)



import torch
from torch import nn

import meddlr.ops as oF
import meddlr.ops.complex as cplx

__all__ = ["SenseModel"]


class SenseModel(nn.Module):
    """
    A module that computes forward and adjoint SENSE operations.

    The forward operation converts a complex image -> multi-coil kspace.
    The adjoint operation converts multi-coil kspace -> a complex image.

    This module also supports multiple sensitivity maps. This is useful if
    you would like to generate images from multiple estimated sensitivity maps.
    This module also works with single coil inputs as long as the #coils dimension
    is set to 1.

    Attributes:
        maps (torch.Tensor): Sensitivity maps. Shape ``(B, H, W, #coils, #maps, [2])``.
        weights (torch.Tensor, optional): Undersampling masks (if applicable).
            Shape ``(B, H, W)`` or ``(B, H, W, #coils, #coils)``.
    """

    def __init__(self, maps: torch.Tensor, weights: torch.Tensor = None):
        """
        Args:
            maps (torch.Tensor): Sensitivity maps.
            weights (torch.Tensor): Undersampling masks.
                If ``None``, it is assumed that inputs are fully-sampled.
        """
        super().__init__()

        self.maps = maps
        if weights is None:
            self.weights = 1.0
        else:
            self.weights = weights

    def _adjoint_op(self, kspace):
        """
        Args:
            kspace: Shape (B,H,W,#coils,[2])
        Returns:
            image: Shape (B,H,W,#maps,[2])
        """
        image = oF.ifft2c(self.weights * kspace, channels_last=True)
        if cplx.is_complex_as_real(kspace):
            image = cplx.mul(image.unsqueeze(-2), cplx.conj(self.maps))  # [B,...,#coils,#maps,2]
            return image.sum(-3)
        else:
            image = cplx.mul(image.unsqueeze(-1), cplx.conj(self.maps))  # [B,...,#coils,#maps,1]
            return image.sum(-2)

    def _forward_op(self, image):
        """
        Args:
            image: Shape (B,H,W,#maps,[2])
        Returns:
            kspace: Shape (B,H,W,#coils,[2])
        """
        if cplx.is_complex_as_real(image):
            kspace = cplx.mul(image.unsqueeze(-3), self.maps)  # [B,...,1,#maps,2]
            kspace = self.weights * oF.fft2c(kspace.sum(-2), channels_last=True)  # [B,...,#coils,2]
        else:
            kspace = cplx.mul(image.unsqueeze(-2), self.maps)
            kspace = self.weights * oF.fft2c(kspace.sum(-1), channels_last=True)
        return kspace

    def forward(self, input: torch.Tensor, adjoint: bool = False):
        """Run forward or adjoint SENSE operation on the input.

        Depending on if ``adjoint=True``, the input should either be the
        k-space or the complex image. The shapes for these are as follows:
            - kspace: ``(B, H, W, #coils, [2])
            - image: ``(B, H, W, #maps, [2])``

        Args:
            input (torch.Tensor): If ``adjoint=True``, this is the multi-coil k-space,
                else it is the image.
            adjoint (bool, optional): If ``True``, use adjoint operation.

        Returns:
            torch.Tensor: If ``adjoint=True``, the image, else multi-coil k-space.
        """
        if adjoint:
            output = self._adjoint_op(input)
        else:
            output = self._forward_op(input)
        return output



import inspect
import pprint
from numbers import Number
from typing import Any, Dict, Sequence

import torch

from meddlr.transforms.mixins import DeviceMixin, TransformCacheMixin
from meddlr.transforms.param_kind import ParamKind
from meddlr.transforms.tf_scheduler import SchedulableMixin

__all__ = ["TransformGen"]


class TransformGen(DeviceMixin, SchedulableMixin, TransformCacheMixin):
    """
    TransformGen takes an array of type float as input.
    It creates a :class:`Transform` based on the given image, sometimes with
    randomness. The transform can then be used to transform images
    or other data (boxes, points, annotations, etc.) associated with it.
    The assumption made in this class is that the image itself is sufficient to
    instantiate a transform. When this assumption is not true, you need to
    create the transforms by your own.
    A list of `TransformGen` can be applied with :func:`apply_transform_gens`.
    """

    def __init__(
        self, params: Dict[str, Any] = None, p: float = 0.0, param_kinds: Dict[str, Any] = None
    ) -> None:
        from meddlr.transforms.tf_scheduler import TFScheduler

        self._params = {}

        if params is None:
            params = {}
        if param_kinds is None:
            param_kinds = {}
        params.update({"p": p})
        self._set_attributes(params)
        self._param_kinds = param_kinds
        self._schedulers: Sequence[TFScheduler] = []

        self._generator = None
        self._device = "cpu"

    def _set_attributes(self, params=None, **kwargs):
        if params is None:
            params = {}
        params.update(kwargs)
        if params:
            self._params.update(
                {k: v for k, v in params.items() if k != "self" and not k.startswith("_")}
            )

    def get_transform(self, input):
        raise NotImplementedError

    def reset(self):
        self._reset_transform()

    def __getattr__(self, name):
        if name in self._params:
            return self._params[name]
        raise AttributeError(f"Attribute '{name}' does not exist in class {type(self)}")

    def _get_param_values(self, use_schedulers=False):
        if not use_schedulers:
            return self._params

        params = self._params.copy()
        for s in self._schedulers:
            params.update(s.get_params())
        return params

    def _rand(self) -> float:
        """Uniform sample between [0, 1) using ``self._generator``.

        Returns:
            float: The sample between [0, 1).
        """
        return torch.rand(1, generator=self._generator).cpu().item()

    def _rand_choice(self, n=None, probs: torch.Tensor = None) -> int:
        """Chooses random integer between [0, n-1].

        Args:
            n (int): Number of choices. This is required if ``probs``
                is not specified.
            probs (torch.Tensor): The probability tensor.

        Returns:
            int: The index of the selected choice.
        """
        device = "cpu" if self._generator is None else self._generator.device
        if probs is None:
            probs = torch.ones(n, device=device) / n
        return torch.multinomial(probs.to(device), 1, generator=self._generator).cpu().item()

    def _rand_range(self, low, high, size: int = None):
        """Uniform float random number between [low, high).

        Args:
            low (number-like): The lower bound.
            high (number-like): The upper bound.
            size (int): Number of samples to draw in the range.

        Returns:
            float: A uniformly sampled number in range [low, high).
        """
        if size is None:
            size = 1

        if low > high:
            high, low = low, high

        if high - low == 0:
            val = low
        else:
            val = (low + (high - low) * torch.rand(size, generator=self._generator)).cpu().item()
        return val

    def _format_param(self, val, kind: ParamKind, ndim=None):
        if kind == ParamKind.MULTI_ARG:
            if isinstance(val, Number):
                return ((-val, val),) * ndim
            elif isinstance(val, (list, tuple)):
                out = []
                if len(val) == 1:
                    val = val * ndim
                for v in val:
                    if isinstance(v, (list, tuple)):
                        out.append(v)
                    elif isinstance(v, Number):
                        out.append((-v, v))
                    else:
                        raise ValueError(f"Type {type(val)} not supported - val={val}")
                return type(val)(out)
        return val

    def seed(self, value: int):
        self._generator = torch.Generator(device=self._device).manual_seed(value)
        return self

    def __repr__(self):
        """
        Produce something like:
        "MyTransformGen(field1={self.field1}, field2={self.field2})"
        """
        try:
            sig = inspect.signature(self.__init__)
            classname = type(self).__name__
            argstr = []
            for name, param in sig.parameters.items():
                assert (
                    param.kind != param.VAR_POSITIONAL and param.kind != param.VAR_KEYWORD
                ), "The default __repr__ doesn't support *args or **kwargs"
                assert hasattr(self, name), (
                    "Attribute {} not found! "
                    "Default __repr__ only works if attributes match "
                    "the constructor.".format(name)
                )
                attr = getattr(self, name)
                default = param.default
                if default is attr:
                    continue
                argstr.append("{}={}".format(name, pprint.pformat(attr)))
            return "{}({})".format(classname, ", ".join(argstr))
        except AssertionError:
            return super().__repr__()

    def __str__(self) -> str:
        return self.__repr__()



import itertools
import logging

import numpy as np
import torch

try:
    from scipy.ndimage.morphology import binary_erosion, distance_transform_edt
except ImportError:  # pragma: no cover
    binary_erosion = None
    distance_transform_edt = None

try:
    from medpy.metric import assd as _assd
except ImportError:  # pragma: no cover
    _assd = None


from meddlr.metrics.functional import util as mFutil

__all__ = [
    "dice_score",
    "volumetric_overlap_error",
    "coefficient_variation",
    "average_symmetric_surface_distance",
    "dice",
    "cv",
    "voe",
    "assd",
]


logger = logging.getLogger(__name__)


def dice_score(y_pred, y_true):
    """Computes dice score coefficient.

    Args:
        y_pred (torch.Tensor): Predicted categorical segmentation.
            Expected shape: ``(B, C, H, ...)``.
        y_true (torch.Tensor): Ground truth categorical segmentation.
            Expected shape: ``(B, C, H, ...)``.

    Returns:
        torch.Tensor: Dice score. Shape: ``(B, C)``.
    """
    is_tensor = isinstance(y_pred, torch.Tensor)

    y_pred = mFutil.to_bool(y_pred)
    y_true = mFutil.to_bool(y_true)
    y_pred, y_true = mFutil.flatten_other_dims((y_pred, y_true), dim=(0, 1))

    count_nonzero = torch.count_nonzero if is_tensor else np.count_nonzero
    size_i1 = count_nonzero(y_pred, -1)
    size_i2 = count_nonzero(y_true, -1)
    intersection = count_nonzero(y_pred & y_true, -1)

    return 2.0 * intersection / (size_i1 + size_i2)


def volumetric_overlap_error(y_pred, y_true):
    """Computes volumetric overlap error.

    Args:
        y_pred (torch.Tensor): Predicted categorical segmentation.
            Expected shape: ``(B, C, H, ...)``.
        y_true (torch.Tensor): Ground truth categorical segmentation.
            Expected shape: ``(B, C, H, ...)``.

    Returns:
        torch.Tensor: Dice score. Shape: ``(B, C)``.
    """
    is_tensor = isinstance(y_pred, torch.Tensor)

    y_pred = mFutil.to_bool(y_pred)
    y_true = mFutil.to_bool(y_true)
    y_pred, y_true = mFutil.flatten_other_dims((y_pred, y_true), dim=(0, 1))

    count_nonzero = torch.count_nonzero if is_tensor else np.count_nonzero
    intersection = count_nonzero(y_true & y_pred, -1)
    union = count_nonzero(y_true | y_pred, -1)
    union = union.type(torch.float) if is_tensor else union.astype(np.float)

    return 1 - intersection / union


def coefficient_variation(y_pred, y_true):
    """Computes coefficient of variation.

    Args:
        y_pred (torch.Tensor): Predicted categorical segmentation.
            Expected shape: ``(B, C, H, ...)``.
        y_true (torch.Tensor): Ground truth categorical segmentation.
            Expected shape: ``(B, C, H, ...)``.

    Returns:
        torch.Tensor: Dice score. Shape: ``(B, C)``.
    """
    is_tensor = isinstance(y_pred, torch.Tensor)

    y_pred = mFutil.to_bool(y_pred)
    y_true = mFutil.to_bool(y_true)
    y_pred, y_true = mFutil.flatten_other_dims((y_pred, y_true), dim=(0, 1))

    count_nonzero = torch.count_nonzero if is_tensor else np.count_nonzero
    size_i1 = count_nonzero(y_pred, -1)
    size_i2 = count_nonzero(y_true, -1)

    if is_tensor:
        cat_tensor = torch.stack([size_i1, size_i2], dim=0).type(torch.float32)
        std = torch.std(cat_tensor, dim=0)
        mean = torch.mean(cat_tensor, dim=0)
    else:
        std = np.std([size_i1, size_i2], axis=0)
        mean = np.mean([size_i1, size_i2], axis=0)

    return std / mean


def average_symmetric_surface_distance(y_pred, y_true, spacing=None, connectivity=1, crop=True):
    """Computes average symmetric surface distance.

    Args:
        y_pred (torch.Tensor): Predicted categorical segmentation.
            Expected shape: ``(B, C, H, ...)``.
        y_true (torch.Tensor): Ground truth categorical segmentation.
            Expected shape: ``(B, C, H, ...)``.

    Returns:
        torch.Tensor: Dice score. Shape: ``(B, C)``.
    """
    if _assd is None:
        raise ModuleNotFoundError(
            "assd requires the medpy package. Please install using `pip install medpy`."
        )

    # TODO: check why this statement is needed.
    if not connectivity:
        connectivity = 1

    is_tensor = isinstance(y_pred, torch.Tensor)
    if is_tensor:
        if y_pred.requires_grad or y_true.requires_grad:
            logger.warning("Average symmetric surface distance does not preserve gradients.")
        y_pred = y_pred.detach().cpu().numpy()
        y_true = y_true.detach().cpu().numpy()

    label = 1
    if y_pred.dtype not in (np.bool, bool):
        y_pred = y_pred == label
    if y_true.dtype not in (np.bool, bool):
        y_true = y_true == label

    B, C = y_pred.shape[:2]
    out_matrix = torch.zeros(B, C) if is_tensor else np.zeros((B, C))
    for b in range(B):
        for c in range(C):
            bc_pred, bc_true = y_pred[b, c], y_true[b, c]
            if crop:
                bc_pred, bc_true = _crop_to_joint_roi(bc_pred, bc_true)

            out_matrix[b, c] = _assd(
                bc_pred, bc_true, voxelspacing=spacing, connectivity=connectivity
            )
    return out_matrix


def _crop_to_joint_roi(y_pred: np.ndarray, y_true: np.ndarray):
    joint_roi = np.asarray(y_pred | y_true)
    ndim = joint_roi.ndim
    if not np.any(joint_roi):
        return y_pred, y_true

    bbox = []
    for ax in itertools.combinations(reversed(range(ndim)), ndim - 1):
        arr = np.any(joint_roi, ax)
        arg_max = np.where(arr == arr.max())[0]
        min_d = max(arg_max[0], 0)
        max_d = arg_max[-1] + 1
        bbox.append(slice(min_d, max_d))
    bbox = tuple(bbox)

    return y_pred[bbox], y_true[bbox]


dice = dice_score
cv = coefficient_variation
voe = volumetric_overlap_error
assd = average_symmetric_surface_distance



"""Compressed Sensing (2D).

This file contains an implementation of the Compressed Sensing
framework by Lustig, et al. using the Python Package sigpy.
See the tutorial for details.

Tutorial:
    https://sigpy.readthedocs.io/en/latest/generated/sigpy.mri.app.L1WaveletRecon.html#sigpy.mri.app.L1WaveletRecon

Reference:
    Lustig, M., Donoho, D., & Pauly, J. M. (2007). Sparse MRI: The application of compressed
    sensing for rapid MR imaging. Magnetic Resonance in Medicine, 58(6), 1082-1195.
"""
import numpy as np
import sigpy as sp
import sigpy.mri as mr
import torch
from torch import nn

import meddlr.ops.complex as cplx
from meddlr.config.config import configurable
from meddlr.utils.general import move_to_device
from meddlr.utils.transforms import SenseModel

from .build import META_ARCH_REGISTRY

try:
    import cupy as cp

    _CUPY_AVAILABLE = True
except ImportError:
    cp = None
    _CUPY_AVAILABLE = False

__all__ = ["CSModel"]


@META_ARCH_REGISTRY.register()
class CSModel(nn.Module):
    @configurable
    def __init__(self, reg: float, max_iter: int, device="cpu", num_emaps: int = 1):
        """
        Args:
            reg (float): The regularization strength.
            max_iter (int): Maximum number of iterations.
            device (str | torch.device, optional): The device to execute on.
            num_emaps (int, optional): Number of estimated sensitivity maps.
                Currently only ``1`` is supported.
        """
        super().__init__()
        if device != torch.device("cpu") and not _CUPY_AVAILABLE:
            raise ModuleNotFoundError(
                f"Requested device {device}, but cupy not installed. "
                f"Install cupy>=9.0 following instructions at "
                f"https://docs.cupy.dev/en/stable/install.html"
            )
        self.device = device

        # Extract network parameters
        self.l1_reg = reg
        self.max_iter = max_iter

        # Data dimensions
        self.num_emaps = num_emaps
        if self.num_emaps != 1:
            raise ValueError("CS currentlyonly supports one sensitivity map.")

    def forward(self, inputs, return_pp=False, vis_training=False):
        """
        TODO: condense into list of dataset dicts.
        Args:
            inputs: Standard ss_recon module input dictionary
                * "kspace": Kspace. If fully sampled, and want to simulate
                    undersampled kspace, provide "mask" argument.
                * "maps": Sensitivity maps
                * "target" (optional): Target image (typically fully sampled).
                * "mask" (optional): Undersampling mask to apply.
                * "signal_model" (optional): The signal model. If provided,
                    "maps" will not be used to estimate the signal model.
                    Use with caution.
            return_pp (bool, optional): If `True`, return post-processing
                parameters "mean", "std", and "norm" if included in the input.
            vis_training (bool, optional): If `True`, force visualize training
                on this pass. Can only be `True` if model is in training mode.

        Returns:
            Dict: A standard ss_recon output dict
                * "pred": The reconstructed image
                * "target" (optional): The target image.
                    Added if provided in the input.
                * "mean"/"std"/"norm" (optional): Pre-processing parameters.
                    Added if provided in the input.
                * "zf_image": The zero-filled image.
                    Added when model is in eval mode.
        """
        if inputs["kspace"].shape[0] != 1:
            raise ValueError("Only batch size == 1 is supported in compressed sensing")

        # Need to fetch device at runtime for proper data transfer.
        # device = self.resnets[0].final_layer.weight.device
        device = self.device
        inputs = move_to_device(inputs, device)
        kspace = inputs["kspace"]
        target = inputs.get("target", None)
        mask = inputs.get("mask", None)
        A = inputs.get("signal_model", None)
        maps = inputs["maps"]
        num_maps_dim = -2 if cplx.is_complex_as_real(maps) else -1
        if self.num_emaps != maps.size()[num_maps_dim]:
            raise ValueError("Incorrect number of ESPIRiT maps! Re-prep data...")

        if mask is None:
            mask = cplx.get_mask(kspace)
        kspace *= mask

        # Declare signal model.
        if A is None:
            A = SenseModel(maps, weights=mask)
        zf_image = A(kspace, adjoint=True)

        # Channel-first - (#coils, ky, kz)
        # TODO: Generalize to 3D
        kspace = kspace[0].permute((2, 0, 1))
        maps = maps.squeeze(num_maps_dim)[0].permute((2, 0, 1))
        mask = mask[0].permute((2, 0, 1))

        xp = np if device == torch.device("cpu") else cp
        kspace = xp.asarray(kspace)
        maps = xp.asarray(maps)
        mask = xp.asarray(mask)

        image = mr.app.L1WaveletRecon(
            kspace,
            maps,
            self.l1_reg,
            weights=mask,
            max_iter=self.max_iter,
            device=sp.get_device(kspace),
        ).run()
        image = torch.as_tensor(image, device=device)
        image = image.unsqueeze(0).unsqueeze(-1)

        output_dict = {"pred": image, "target": target}  # N x Y x Z x 1 x 2  # N x Y x Z x 1 x 2
        if return_pp:
            output_dict.update({k: inputs[k] for k in ["mean", "std", "norm"]})

        if not self.training:
            output_dict["zf_image"] = zf_image

        return output_dict

    @classmethod
    def from_config(cls, cfg):
        return {
            "reg": cfg.MODEL.CS.REGULARIZATION,
            "max_iter": cfg.MODEL.CS.MAX_ITER,
            "device": cfg.MODEL.DEVICE,
            "num_emaps": cfg.MODEL.UNROLLED.NUM_EMAPS,
        }



import itertools
import logging
import random
from collections import defaultdict
from typing import Dict, Mapping, Sequence, Tuple, Union

import numpy as np
from torch.utils.data import DataLoader

from meddlr.data.catalog import DatasetCatalog
from meddlr.data.collate import collate_by_supervision, default_collate
from meddlr.data.samplers.build import build_train_sampler, build_val_sampler
from meddlr.data.slice_dataset import SliceData
from meddlr.data.transforms import transform as T
from meddlr.data.transforms.subsample import build_mask_func

__all__ = ["build_recon_train_loader", "build_recon_val_loader"]


def get_recon_dataset_dicts(
    dataset_names: Sequence[str],
    num_scans_total: Union[int, Tuple] = -1,
    num_scans_subsample: int = 0,
    seed: int = 1000,
    accelerations=(),
    filter_by=(),
):
    """Get recon datasets and perform filtering.

    Given the same seed, scans will be selected in a fixed order. For example,
    let ``num_scans_total=3`` in experiment A and ``num_scans_total=4`` in
    experiment B. If the seed is the same for both experiments, then the scans
    selected in experiment A will be a subset of the scans selected in
    experiment B. This is to simulate data as a growing set- we can never
    "lose" data, only add to the existing set.

    Args:
        dataset_names (str(s)): Datasets to load. See ss_recon/data/datasets/builtin.py
            for built-in datasets.
        num_scans_total (int): Number of total scans to return. If ``-1``, ignored.
        num_scans_subsample (int): Number of scans to mark as only undersampled
            (i.e. unsupervised) scans. These scans will not have a ground truth scan.
        seed (int): The deterministic seed for filtering which scans to select.
            For reproducibility, this seed should be set.
        accelerations (Tuple[float, float], optional): The range of accelerations
            for this dataset. The maximum in the range will be used for retrospective
            undersampling of the unsupervised subset of scans.
        filter_by (Tuple): Metadata key to filter on and the value(s) to filter by.
            If the value for a particular key is a list, any one of the elements in
            the list is a valid entry.
            The schema is ``((k1, v1), (k2, v2), (k3, v3), ...)``.

    Returns:
        List[Dict]: The dataset dictionaries.
    """
    assert len(dataset_names)
    dataset_dicts = [DatasetCatalog.get(dataset_name) for dataset_name in dataset_names]
    dataset_dicts = list(itertools.chain.from_iterable(dataset_dicts))
    num_before = len(dataset_dicts)

    logger = logging.getLogger(__name__)

    state = random.Random(seed)
    limit_scans = (
        num_scans_total > 0 if isinstance(num_scans_total, (float, int)) else bool(num_scans_total)
    )
    if limit_scans or num_scans_subsample > 0:
        # Sort to ensure same order for all users.
        # Shuffle for randomness.
        dataset_dicts = sorted(dataset_dicts, key=lambda x: x["file_name"])
        state.shuffle(dataset_dicts)

    if filter_by:
        for k, v in filter_by:
            num_before = len(dataset_dicts)
            if not isinstance(v, (list, tuple)):
                v = (v,)
            else:
                # This helps us ignore casting differences between
                # list and tuple.
                extra_v = []
                for _v in v:
                    if isinstance(_v, list):
                        extra_v.append(tuple(_v))
                    elif isinstance(_v, tuple):
                        extra_v.append(list(_v))
                v = list(v) + extra_v
            dataset_dicts = [
                dd
                for dd in dataset_dicts
                if dd.get(k, None) in v or dd.get("_metadata", {}).get(k, None) in v
            ]
            num_after = len(dataset_dicts)
            logger.info(
                f"Filtered by {k}: Dropped {num_before - num_after} scans. "
                f"{num_after} scans remaining"
            )

    num_after_filter = len(dataset_dicts)
    if isinstance(num_scans_total, int) and num_scans_total > 0:
        dataset_dicts = dataset_dicts[:num_scans_total]
    elif isinstance(num_scans_total, Tuple) and num_scans_total:
        dataset_dicts = _limit_data_by_group(dataset_dicts, num_scans_total)

    num_after = len(dataset_dicts)
    logger.info(
        "Dropped {} scans. {} scans remaining".format(num_after_filter - num_after, num_after)
    )

    num_scans_subsample = max(0, num_scans_subsample)
    if num_scans_subsample > 0:
        if num_scans_subsample > len(dataset_dicts):
            raise ValueError("")
        for dd in dataset_dicts[:num_scans_subsample]:
            dd["_is_unsupervised"] = True
            # Select the maximum acceleration when doing undersampling.
            dd["_acceleration"] = max(accelerations)
    logger.info(
        "Dropped references for {}/{} scans. "
        "{} scans with reference remaining".format(
            num_scans_subsample, num_after, num_after - num_scans_subsample
        )
    )

    return dataset_dicts


def _limit_data_by_group(dataset_dicts, num_scans_total: Tuple[str, Dict]):
    if len(num_scans_total) > 1:
        raise ValueError("Currently can only limit scans by one field")
    metadata_key, cat_to_limit = num_scans_total[0]
    if not isinstance(cat_to_limit, Mapping):
        if not isinstance(cat_to_limit, Sequence) or len(cat_to_limit) % 2 != 0:
            raise ValueError(
                "Categories and limits must be provided as a dictionary or even-length tuple. "
                "The tuple should be (group1, limit1, group2, limit2, ...)"
            )
        cat_to_limit = {k: v for k, v in zip(cat_to_limit[::2], cat_to_limit[1::2])}
    cat_to_num = defaultdict(int)
    new_dataset_dicts = []
    for dd in dataset_dicts:
        metadata_val = dd.get(metadata_key, dd.get("_metadata", {}).get(metadata_key, None))
        for k in cat_to_limit:
            if isinstance(k, (tuple, list)) and metadata_val in k:
                metadata_val = k
                break
        if metadata_val is None or metadata_val not in cat_to_limit:
            new_dataset_dicts.append(dd)
            continue
        if cat_to_num.get(metadata_val, 0) >= cat_to_limit[metadata_val]:
            # Quota for this metadata field reached.
            continue
        new_dataset_dicts.append(dd)
        cat_to_num[metadata_val] += 1
    return new_dataset_dicts


def _build_dataset(cfg, dataset_dicts, data_transform, dataset_type=None, is_eval=False, **kwargs):
    keys = cfg.DATALOADER.DATA_KEYS
    if keys:
        assert all(
            len(x) == 2 for x in keys
        ), "cfg.DATALOADER.DATA_KEYS should be sequence of tuples of len 2"
        keys = {k: v for k, v in keys}

    if dataset_type is None:
        dataset_type = SliceData
    return dataset_type(
        dataset_dicts, data_transform, keys=keys, include_metadata=is_eval, **kwargs
    )


def _get_default_dataset_type(dataset_name):
    """Returns the default dataset type based on the dataset name.

    TODO: This function and its call hierarchy need to be refactored.
    """
    return SliceData


def build_recon_train_loader(cfg, dataset_type=None):
    if (
        cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_TOTAL > 0
        and cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_TOTAL_BY_GROUP
    ):
        raise ValueError(
            "`DATALOADER.SUBSAMPLE_TRAIN.NUM_TOTAL` and "
            "`DATALOADER.SUBSAMPLE_TRAIN.NUM_TOTAL_BY_GROUP` are mutually exclusive."
        )
    num_scans_total = (
        cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_TOTAL_BY_GROUP
        if cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_TOTAL_BY_GROUP
        else cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_TOTAL
    )

    dataset_dicts = get_recon_dataset_dicts(
        dataset_names=cfg.DATASETS.TRAIN,
        num_scans_total=num_scans_total,
        num_scans_subsample=cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_UNDERSAMPLED,
        seed=cfg.DATALOADER.SUBSAMPLE_TRAIN.SEED,
        accelerations=cfg.AUG_TRAIN.UNDERSAMPLE.ACCELERATIONS,
        filter_by=cfg.DATALOADER.FILTER.BY,
    )
    if dataset_type is None:
        dataset_type = _get_default_dataset_type(cfg.DATASETS.TRAIN[0])

    mask_func = build_mask_func(cfg.AUG_TRAIN)
    data_transform = T.DataTransform(
        cfg,
        mask_func,
        is_test=False,
        add_noise=cfg.AUG_TRAIN.USE_NOISE,
        add_motion=cfg.AUG_TRAIN.USE_MOTION,
    )

    train_data = _build_dataset(cfg, dataset_dicts, data_transform, dataset_type)
    is_semi_supervised = len(train_data.get_unsupervised_idxs()) > 0
    collate_fn = collate_by_supervision if is_semi_supervised else default_collate

    # Build sampler.
    sampler, is_batch_sampler = build_train_sampler(cfg, train_data)
    shuffle = not sampler  # shuffling should be handled by sampler, if specified.
    if is_batch_sampler:
        dl_kwargs = {"batch_sampler": sampler}
    else:
        dl_kwargs = {
            "sampler": sampler,
            "batch_size": cfg.SOLVER.TRAIN_BATCH_SIZE,
            "shuffle": shuffle,
            "drop_last": cfg.DATALOADER.DROP_LAST,
        }

    train_loader = DataLoader(
        dataset=train_data,
        num_workers=cfg.DATALOADER.NUM_WORKERS,
        pin_memory=True,
        collate_fn=collate_fn,
        prefetch_factor=cfg.DATALOADER.PREFETCH_FACTOR,
        **dl_kwargs,
    )
    return train_loader


def build_recon_val_loader(
    cfg,
    dataset_name,
    as_test: bool = False,
    add_noise: bool = False,
    add_motion: bool = False,
    dataset_type=None,
):
    if (
        cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_VAL > 0
        and cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_VAL_BY_GROUP
    ):
        raise ValueError(
            "`DATALOADER.SUBSAMPLE_TRAIN.NUM_VAL` and "
            "`DATALOADER.SUBSAMPLE_TRAIN.NUM_VAL_BY_GROUP` are mutually exclusive."
        )
    num_scans_total = (
        cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_VAL_BY_GROUP
        if cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_VAL_BY_GROUP
        else cfg.DATALOADER.SUBSAMPLE_TRAIN.NUM_VAL
    )

    dataset_dicts = get_recon_dataset_dicts(
        dataset_names=[dataset_name],
        filter_by=cfg.DATALOADER.FILTER.BY,
        num_scans_total=num_scans_total,
        seed=cfg.DATALOADER.SUBSAMPLE_TRAIN.SEED,
    )
    if dataset_type is None:
        dataset_type = _get_default_dataset_type(dataset_name)

    mask_func = build_mask_func(cfg.AUG_TRAIN)
    data_transform = T.DataTransform(
        cfg, mask_func, is_test=as_test, add_noise=add_noise, add_motion=add_motion
    )

    val_data = _build_dataset(
        cfg, dataset_dicts, data_transform, is_eval=True, dataset_type=dataset_type
    )

    # Build sampler.
    sampler, is_batch_sampler = build_val_sampler(cfg, val_data)
    if is_batch_sampler:
        dl_kwargs = {"batch_sampler": sampler}
    else:
        dl_kwargs = {
            "sampler": sampler,
            "batch_size": cfg.SOLVER.TEST_BATCH_SIZE,
            "shuffle": False,
            "drop_last": False,
        }

    val_loader = DataLoader(
        dataset=val_data,
        num_workers=cfg.DATALOADER.NUM_WORKERS,
        pin_memory=True,
        collate_fn=default_collate,
        prefetch_factor=cfg.DATALOADER.PREFETCH_FACTOR,
        **dl_kwargs,
    )
    return val_loader


def build_data_loaders_per_scan(cfg, dataset_name, accelerations=None):
    """Creates a data loader for each unique scan.

    TODO: Deprecate this function and incorporate scan action into standard
    testing pipeline
    """
    dataset_dicts = get_recon_dataset_dicts(dataset_names=[dataset_name])

    if accelerations is None:
        accelerations = cfg.AUG_TRAIN.UNDERSAMPLE.ACCELERATIONS
        accelerations = list(np.arange(accelerations[0], accelerations[1]))

    assert (
        len(cfg.AUG_TRAIN.UNDERSAMPLE.CENTER_FRACTIONS) <= 1
    ), "Currently only support single center fraction during testing"

    loaders = defaultdict(dict)
    for acc in accelerations:
        aug_cfg = cfg.AUG_TRAIN.clone()
        aug_cfg.defrost()
        aug_cfg.UNDERSAMPLE.ACCELERATIONS = (acc,)
        aug_cfg.freeze()
        for dataset_dict in dataset_dicts:
            mask_func = build_mask_func(aug_cfg)
            data_transform = T.DataTransform(cfg, mask_func, is_test=True)
            train_data = _build_dataset(cfg, [dataset_dict], data_transform, is_eval=True)
            loader = DataLoader(
                dataset=train_data,
                batch_size=cfg.SOLVER.TEST_BATCH_SIZE,
                shuffle=False,
                num_workers=cfg.DATALOADER.NUM_WORKERS,
                drop_last=False,
                pin_memory=True,
                collate_fn=default_collate,
            )
            loaders["{}x".format(acc)][dataset_dict["file_name"]] = loader

    return loaders



import math
from bisect import bisect_right
from typing import List

import torch

# NOTE: PyTorch's LR scheduler interface uses names that assume the LR changes
# only on epoch boundaries. We typically use iteration based schedules instead.
# As a result, "epoch" (e.g., as in self.last_epoch) should be understood to
# mean "iteration" instead.


class NoOpLR(torch.optim.lr_scheduler._LRScheduler):
    def __init__(self, optimizer, last_epoch=-1):
        super().__init__(optimizer, last_epoch)

    def get_lr(self) -> List[float]:
        return self.base_lrs

    def _compute_values(self) -> List[float]:
        # The new interface
        return self.get_lr()


class WarmupMultiStepLR(torch.optim.lr_scheduler._LRScheduler):
    """

    Adapted from
    https://github.com/facebookresearch/detectron2
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        milestones: List[int],
        gamma: float = 0.1,
        warmup_factor: float = 0.001,
        warmup_iters: int = 1000,
        warmup_method: str = "linear",
        last_epoch: int = -1,
    ):
        if not list(milestones) == sorted(milestones):
            raise ValueError(
                "Milestones should be a list of" " increasing integers. Got {}", milestones
            )
        self.milestones = milestones
        self.gamma = gamma
        self.warmup_factor = warmup_factor
        self.warmup_iters = warmup_iters
        self.warmup_method = warmup_method
        super().__init__(optimizer, last_epoch)

    def get_lr(self) -> List[float]:
        warmup_factor = _get_warmup_factor_at_iter(
            self.warmup_method, self.last_epoch, self.warmup_iters, self.warmup_factor
        )
        return [
            base_lr * warmup_factor * self.gamma ** bisect_right(self.milestones, self.last_epoch)
            for base_lr in self.base_lrs
        ]

    def _compute_values(self) -> List[float]:
        # The new interface
        return self.get_lr()


class WarmupCosineLR(torch.optim.lr_scheduler._LRScheduler):
    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        max_iters: int,
        warmup_factor: float = 0.001,
        warmup_iters: int = 1000,
        warmup_method: str = "linear",
        last_epoch: int = -1,
    ):
        """
        Adapted from
        https://github.com/facebookresearch/detectron2
        """
        self.max_iters = max_iters
        self.warmup_factor = warmup_factor
        self.warmup_iters = warmup_iters
        self.warmup_method = warmup_method
        super().__init__(optimizer, last_epoch)

    def get_lr(self) -> List[float]:
        warmup_factor = _get_warmup_factor_at_iter(
            self.warmup_method, self.last_epoch, self.warmup_iters, self.warmup_factor
        )
        # Different definitions of half-cosine with warmup are possible. For
        # simplicity we multiply the standard half-cosine schedule by the warmup
        # factor. An alternative is to start the period of the cosine at warmup
        # iters instead of at 0. In the case that warmup_iters << max_iters the
        # two are very close to each other.
        return [
            base_lr
            * warmup_factor
            * 0.5
            * (1.0 + math.cos(math.pi * self.last_epoch / self.max_iters))
            for base_lr in self.base_lrs
        ]

    def _compute_values(self) -> List[float]:
        # The new interface
        return self.get_lr()


def _get_warmup_factor_at_iter(
    method: str, iter: int, warmup_iters: int, warmup_factor: float
) -> float:
    """
    Return the learning rate warmup factor at a specific iteration.
    See https://arxiv.org/abs/1706.02677 for more details.

    Args:
        method (str): warmup method; either "constant" or "linear".
        iter (int): iteration at which to calculate the warmup factor.
        warmup_iters (int): the number of warmup iterations.
        warmup_factor (float): the base warmup factor (the meaning changes
            according to the method used).

    Returns:
        float: the effective warmup factor at the given iteration.
    """
    if iter >= warmup_iters:
        return 1.0

    if method == "constant":
        return warmup_factor
    elif method == "linear":
        alpha = iter / warmup_iters
        return warmup_factor * (1 - alpha) + alpha
    else:
        raise ValueError("Unknown warmup method: {}".format(method))



import inspect
from abc import ABC, abstractmethod
from typing import Tuple, Union

import torch
import torch.nn as nn

from meddlr.modeling.blocks.conv_blocks import SimpleConvBlockNd

__all__ = [
    "ResBlockNd",
    "ResBlockNd",
    "ResBlock3d",
    "ConcatBlockNd",
    "ConcatBlock2d",
    "ConcatBlock3d",
]


class _SimpleFuseBlockNd(nn.Module, ABC):
    """Series of :class:`SimpleConvBlockNd` with residual connection."""

    # Assumes order is the last argument in SimpleConvBlockNd
    _DEFAULT_CONV_BLOCK_ORDER = inspect.getfullargspec(SimpleConvBlockNd).defaults[-1]

    def __init__(
        self,
        in_channels: int,
        kernel_size: Union[int, Tuple[int, ...]],
        n_blocks: int,
        dimension: int,
        connect_before: str = None,
        **kwargs,
    ):
        """
        Args:
            in_channels (int): Number of channels in the input.
            kernel_size (`int(s)`): Convolution kernel size.
            n_blocks (int): Number of conv blocks.
            dimension (int): Integer specifying the dimension of convolution.
            connect_before (str, optional): Layer to add residual connection before in conv block.
                For example, if `n_blocks=1`, conv block `order=("conv", "batchnorm", "relu")`,
                and `connect_before="relu"`, residual block will look like below. If `None`,
                residual connection will be made after full conv block.
                    x -> Conv -> BatchNorm -> + -> ReLU
                    |                         ^
                    |                         |
                    --------------------------
            kwargs: `SimpleConvBlockNd` arguments. `in_channels` and `out_channels` required.
        """
        super().__init__()
        self.n_blocks = n_blocks

        # Determine order for connecting before.
        order = kwargs.pop("order", self._DEFAULT_CONV_BLOCK_ORDER)
        if connect_before:
            if connect_before not in order:
                raise ValueError(f"Layer {connect_before} not in conv block `order` ({order})")
            layer_idx = order.index(connect_before)
            if layer_idx == 0:
                raise ValueError(
                    f"Layer {connect_before} occurs first in conv block `order` ({order}). "
                    f"Reduce n_block by 1, set `connect_before=None`, and add `SimpleConvBlockNd` "
                    f"after this residual block."
                )
            standard_order = [order] * (self.n_blocks - 1)
            split_order = [order[:layer_idx], order[layer_idx:]]
            conv_orders = standard_order + split_order
        else:
            conv_orders = [order] * self.n_blocks
        self.connect_before = connect_before

        # Build conv blocks
        self.blocks = nn.ModuleDict(
            {
                f"block_{i+1}": SimpleConvBlockNd(
                    in_channels=in_channels,
                    out_channels=in_channels,
                    kernel_size=kernel_size,
                    dimension=dimension,
                    order=order,
                    **kwargs,
                )
                for i, order in enumerate(conv_orders)
            }
        )
        assert (self.connect_before and len(self.blocks) == self.n_blocks + 1) or (
            not self.connect_before and len(self.blocks) == self.n_blocks
        )

    def forward(self, x):
        out = x
        for i in range(self.n_blocks):
            out = self.blocks[f"block_{i+1}"](out)

        out = self.fuse(x, out)

        # Handle any remaining layers when connect_before is specified
        if self.connect_before:
            out = self.blocks[f"block_{self.n_blocks + 1}"](out)

        return out

    @abstractmethod
    def fuse(self, x, y):
        """Fuse two tensors"""
        pass


class ResBlockNd(_SimpleFuseBlockNd):
    """
    Args:
        in_channels (int): Number of channels in the input.
        kernel_size (`int(s)`): Convolution kernel size.
        n_blocks (int): Number of conv blocks.
        dimension (int): Integer specifying the dimension of convolution.
        connect_before (str, optional): Layer to add residual connection before in conv block.
            For example, if `n_blocks=1`, conv block `order=("conv", "batchnorm", "relu")`,
            and `connect_before="relu"`, residual block will look like below. If `None`,
            residual connection will be made after full conv block.
                x -> Conv -> BatchNorm -> + -> ReLU
                |                         ^
                |                         |
                --------------------------
        kwargs: `SimpleConvBlockNd` arguments. `in_channels` and `out_channels` required.
    """

    def __init__(
        self,
        in_channels: int,
        kernel_size: Union[int, Tuple[int, ...]],
        n_blocks: int,
        dimension: int,
        connect_before: str = None,
        **kwargs,
    ):
        super().__init__(in_channels, kernel_size, n_blocks, dimension, connect_before, **kwargs)

    def fuse(self, x, y):
        return x + y


class ResBlock2d(ResBlockNd):
    """2D implementation of :class:`SimpleResBlock`.

    Args:
        in_channels (int): Number of channels in the input.
        kernel_size (`int(s)`): Convolution kernel size.
        n_blocks (int): Number of conv blocks.
        connect_before (str, optional): Layer to add residual connection before in conv block.
            For example, if `n_blocks=1`, conv block `order=("conv", "batchnorm", "relu")`,
            and `connect_before="relu"`, residual block will look like below. If `None`,
            residual connection will be made after full conv block.
                x -> Conv -> BatchNorm -> + -> ReLU
                |                         ^
                |                         |
                --------------------------
        kwargs: `SimpleConvBlockNd` arguments. `in_channels` and `out_channels` required.
    """

    def __init__(
        self,
        in_channels: int,
        kernel_size: Union[int, Tuple[int, ...]],
        n_blocks: int,
        connect_before: str = None,
        **kwargs,
    ):
        super().__init__(in_channels, kernel_size, n_blocks, 2, connect_before, **kwargs)


class ResBlock3d(ResBlockNd):
    """3D implementation of :class:`SimpleResBlock`.

    Args:
        in_channels (int): Number of channels in the input.
        kernel_size (`int(s)`): Convolution kernel size.
        n_blocks (int): Number of conv blocks
        connect_before (str, optional): Layer to add residual connection before in conv block.
            For example, if `n_blocks=1`, conv block `order=("conv", "batchnorm", "relu")`,
            and `connect_before="relu"`, residual block will look like below. If `None`,
            residual connection will be made after full conv block.
                x -> Conv -> BatchNorm -> + -> ReLU
                |                         ^
                |                         |
                --------------------------
        kwargs: `SimpleConvBlockNd` arguments. `in_channels` and `out_channels` required.
    """

    def __init__(
        self,
        in_channels: int,
        kernel_size: Union[int, Tuple[int, ...]],
        n_blocks: int,
        connect_before: str = None,
        **kwargs,
    ):
        super().__init__(in_channels, kernel_size, n_blocks, 3, connect_before, **kwargs)


class ConcatBlockNd(_SimpleFuseBlockNd):
    """
    Args:
        in_channels (int): Number of channels in the input.
        kernel_size (`int(s)`): Convolution kernel size.
        n_blocks (int): Number of conv blocks.
        dimension (int): Integer specifying the dimension of convolution.
        connect_before (str, optional): Layer to add residual connection before in conv block.
            For example, if `n_blocks=1`, conv block `order=("conv", "batchnorm", "relu")`,
            and `connect_before="relu"`, residual block will look like below. If `None`,
            residual connection will be made after full conv block.
                x -> Conv -> BatchNorm -> [concat] -> ReLU
                |                            ^
                |                            |
                ------------------------------
        kwargs: `SimpleConvBlockNd` arguments. `in_channels` and `out_channels` required.
    """

    def __init__(
        self,
        in_channels: int,
        kernel_size: Union[int, Tuple[int, ...]],
        n_blocks: int,
        dimension: int,
        connect_before: str = None,
        **kwargs,
    ):
        super().__init__(in_channels, kernel_size, n_blocks, dimension, connect_before, **kwargs)

    def fuse(self, x, y):
        return torch.cat([x, y], dim=1)


class ConcatBlock2d(ConcatBlockNd):
    def __init__(
        self,
        in_channels: int,
        kernel_size: Union[int, Tuple[int, ...]],
        n_blocks: int,
        connect_before: str = None,
        **kwargs,
    ):
        super().__init__(in_channels, kernel_size, n_blocks, 2, connect_before, **kwargs)


class ConcatBlock3d(ConcatBlockNd):
    def __init__(
        self,
        in_channels: int,
        kernel_size: Union[int, Tuple[int, ...]],
        n_blocks: int,
        connect_before: str = None,
        **kwargs,
    ):
        super().__init__(in_channels, kernel_size, n_blocks, 3, connect_before, **kwargs)



from typing import Sequence, Tuple, Union

import torch
from torch import nn

from meddlr.config.config import configurable
from meddlr.modeling.blocks import SimpleConvBlockNd
from meddlr.modeling.layers.build import get_layer_kind, get_layer_type
from meddlr.modeling.meta_arch import META_ARCH_REGISTRY


@META_ARCH_REGISTRY.register()
class GeneralizedUNet(nn.Module):
    _VERSION = 1

    @configurable
    def __init__(
        self,
        dimensions,
        in_channels,
        out_channels,
        channels: Sequence[int],
        strides: Sequence[int] = 1,
        kernel_size: Union[Sequence[int], int] = 3,
        up_kernel_size: Union[Sequence[int], int] = None,
        dropout: float = 0.0,
        block_order: Tuple[str, ...] = ("conv", "relu", "conv", "relu", "batchnorm", "dropout"),
    ):
        super().__init__()

        channels = list(channels)
        depth = len(channels)
        assert depth >= 2  # must have at least 2 blocks for U-Net

        kernel_size = self._arg_to_seq(kernel_size, depth)
        up_kernel_size = (
            self._arg_to_seq(up_kernel_size, depth - 1) if up_kernel_size else kernel_size[:-1]
        )
        strides = self._arg_to_seq(strides, depth)
        pool_type = get_layer_type("maxpool", dimension=dimensions)

        block_order_names: Sequence[str] = [x if isinstance(x, str) else x[0] for x in block_order]
        act_idx = [i for i, x in enumerate(block_order_names) if get_layer_kind(x) == "act"][0]
        norm_idx = [i for i, x in enumerate(block_order_names) if get_layer_kind(x) == "norm"][0]
        up_block_order = ("convtranspose", block_order[act_idx], block_order[norm_idx])

        # Down blocks + Bottleneck
        down_blocks = {}
        pool_blocks = {}
        for i, inc, outc, ks, s in zip(
            range(depth), [in_channels] + channels[:-1], channels, kernel_size, strides
        ):
            d_block = SimpleConvBlockNd(
                inc,
                outc,
                kernel_size=ks,
                dimension=dimensions,
                stride=s,
                dropout=dropout,
                order=block_order,
            )
            down_blocks[f"block{i}"] = d_block
            if i < depth - 1:
                pool_blocks[f"block{i}"] = pool_type(2)
        self.down_blocks = nn.ModuleDict(down_blocks)
        self.pool_blocks = nn.ModuleDict(pool_blocks)

        # Up blocks - Up conv + concat + Block
        up_blocks = {}
        for i, inc, outc, ks, s in zip(
            range(depth - 2, -1, -1),
            channels[::-1],
            channels[::-1][1:],
            kernel_size[::-1],
            strides[::-1],
        ):
            conv_t = SimpleConvBlockNd(
                inc,
                outc,
                dimension=dimensions,
                kernel_size=2,
                stride=2,
                order=up_block_order,
                padding=None,
            )
            block = SimpleConvBlockNd(
                outc * 2,
                outc,
                kernel_size=ks,
                dimension=dimensions,
                stride=s,
                dropout=dropout,
                order=block_order,
            )
            block = nn.ModuleList([conv_t, block])
            up_blocks[f"block{i}"] = block
        self.up_blocks = nn.ModuleDict(up_blocks)

        self.output_block = get_layer_type("conv", dimension=dimensions)(
            channels[0],
            out_channels,
            kernel_size=1,
            stride=1,
            padding=0,
        )

    @property
    def bottleneck(self):
        """Easy access property for bottleneck layer."""
        return self.down_blocks[list(self.down_blocks.keys())[-1]]

    @property
    def depth(self):
        return len(self.down_blocks)

    def forward(self, x):
        down_blocks = self.down_blocks
        pool_blocks = self.pool_blocks
        up_blocks = self.up_blocks

        skip_connections = []
        for i in range(len(down_blocks)):
            x = self.down_blocks[f"block{i}"](x)
            if i < len(down_blocks) - 1:
                skip_connections.append(x)
                x = pool_blocks[f"block{i}"](x)

        for i, sc in zip(range(self.depth - 2, -1, -1), skip_connections[::-1]):
            upsample, conv_block = tuple(up_blocks[f"block{i}"])
            x = torch.cat([upsample(x), sc], dim=1)
            x = conv_block(x)

        return self.output_block(x)

    @staticmethod
    def _arg_to_seq(arg, num: int):
        """Converts arguments that are supposed to be sequences of a particular length."""
        if isinstance(arg, Sequence) and len(arg) != num:
            raise ValueError(f"Got `arg` {arg}. Expected len {num}")
        if isinstance(arg, Sequence) and len(arg) == 1:
            arg = arg[0]
        if not isinstance(arg, Sequence):
            arg = (arg,) * num
        return arg

    @classmethod
    def from_config(cls, cfg, **kwargs):
        # Returns kwargs to be passed to __init__
        if "MODEL" in cfg:
            cfg = cfg.MODEL.UNET
        elif "UNET" in cfg:
            cfg = cfg.UNET
        in_channels = cfg.get("IN_CHANNELS", None)
        out_channels = cfg.get("OUT_CHANNELS", None)
        dimensions = kwargs.get("dimensions", 2)
        num_pool_layers = cfg.NUM_POOL_LAYERS
        num_channels = tuple(cfg.CHANNELS * (2 ** i) for i in range(num_pool_layers + 1))

        init_args = {
            "dimensions": dimensions,
            "in_channels": in_channels,
            "out_channels": out_channels,
            "channels": num_channels,
            "dropout": cfg.DROPOUT,
        }
        block_order = cfg.get("BLOCK_ORDER", None)
        if block_order is not None:
            init_args["block_order"] = block_order
        init_args.update(**kwargs)
        return init_args



from numbers import Number
from typing import Dict, Sequence, Union

import meddlr.ops.complex as cplx
from meddlr.data.transforms.transform import Normalizer
from meddlr.evaluation.testing import flatten_results_dict
from meddlr.forward import SenseModel
from meddlr.transforms.build import (
    build_iter_func,
    build_scheduler,
    build_transforms,
    seed_tfm_gens,
)
from meddlr.transforms.gen import RandomTransformChoice
from meddlr.transforms.mixins import DeviceMixin, GeometricMixin
from meddlr.transforms.tf_scheduler import SchedulableMixin, TFScheduler
from meddlr.transforms.transform import NoOpTransform, Transform, TransformList
from meddlr.transforms.transform_gen import TransformGen
from meddlr.utils import env


class MRIReconAugmentor(DeviceMixin):
    """
    The class that manages the organization, generation, and application
    of deterministic and random transforms for MRI reconstruction.
    """

    def __init__(
        self,
        tfms_or_gens: Sequence[Union[Transform, TransformGen]],
        aug_sensitivity_maps: bool = True,
        seed: int = None,
        device=None,
    ) -> None:
        """
        Args:
            aug_sensitivity_maps (bool, optional): If ``True``, apply equivariant,
                image-based transforms to the sensivitiy map.
        """
        if isinstance(tfms_or_gens, TransformList):
            tfms_or_gens = tfms_or_gens.transforms
        self.tfms_or_gens = tfms_or_gens
        self.aug_sensitivity_maps = aug_sensitivity_maps

        if device is not None:
            self.to(device)
        if seed is None:
            seed_tfm_gens(self.tfms_or_gens, seed=seed)

    def __call__(
        self,
        kspace,
        maps=None,
        target=None,
        normalizer: Normalizer = None,
        mask=None,
        mask_gen=None,
        skip_tfm=False,
    ):
        if skip_tfm:
            tfms_equivariant, tfms_invariant = [], []
        else:
            # For now, we assume that transforms generated by RandomTransformChoice
            # return random transforms of the same type (equivariant or invariant).
            # We don't have to filter these transforms as a result.
            transform_gens = [
                x.get_transform() if isinstance(x, RandomTransformChoice) else x
                for x in self.tfms_or_gens
            ]
            if any(isinstance(x, (list, tuple)) for x in transform_gens):
                _temp = []
                for t in transform_gens:
                    if isinstance(t, (list, tuple)):
                        _temp.extend(t)
                    else:
                        _temp.append(t)
                transform_gens = _temp

            tfms_equivariant, tfms_invariant = self._classify_transforms(transform_gens)

        use_img = normalizer is not None or len(tfms_equivariant) > 0

        # Apply equivariant transforms to the SENSE reconstructed image.
        # Note, RSS reconstruction is not currently supported.
        if use_img:
            if mask is True:
                mask = cplx.get_mask(kspace)
            A = SenseModel(maps, weights=mask)
            img = A(kspace, adjoint=True)

        if len(tfms_equivariant) > 0:
            img, target, maps = self._permute_data(img, target, maps, spatial_last=True)
            img, target, maps, tfms_equivariant = self._apply_te(
                tfms_equivariant, img, target, maps
            )
            img, target, maps = self._permute_data(img, target, maps, spatial_last=False)

        if len(tfms_equivariant) > 0:
            A = SenseModel(maps)
            kspace = A(img)

        if mask_gen is not None:
            kspace, mask = mask_gen(kspace)
            img = SenseModel(maps, weights=mask)(kspace, adjoint=True)

        if normalizer:
            normalized = normalizer.normalize(
                **{"masked_kspace": kspace, "image": img, "target": target, "mask": mask}
            )
            kspace = normalized["masked_kspace"]
            target = normalized["target"]
            mean = normalized["mean"]
            std = normalized["std"]
        else:
            mean, std = None, None

        # Apply invariant transforms.
        if len(tfms_invariant) > 0:
            kspace = self._permute_data(kspace, spatial_last=True)
            kspace, tfms_invariant = self._apply_ti(tfms_invariant, kspace)
            kspace = self._permute_data(kspace, spatial_last=False)

        out = {"kspace": kspace, "maps": maps, "target": target, "mean": mean, "std": std}

        for s in self.schedulers():
            s.step(kspace.shape[0])

        return out, tfms_equivariant, tfms_invariant

    def schedulers(self):
        schedulers = [
            tfm.schedulers() for tfm in self.tfms_or_gens if isinstance(tfm, SchedulableMixin)
        ]
        return [x for y in schedulers for x in y]

    def get_tfm_gen_params(self, scalars_only: bool = True):
        """Get dictionary of scheduler parameters."""
        schedulers: Dict[str, Sequence[TFScheduler]] = {
            type(tfm).__name__: tfm._get_param_values(use_schedulers=True)
            for tfm in self.tfms_or_gens
            if isinstance(tfm, SchedulableMixin)
        }
        params = {}
        for tfm_name, p in schedulers.items():
            p = flatten_results_dict(p, delimiter=".")
            # Filter out values that are not scalars
            p = {f"{tfm_name}/{k}": v for k, v in p.items()}
            if scalars_only:
                p = {k: v for k, v in p.items() if isinstance(v, Number)}
            params.update(p)
        return params

    def _classify_transforms(self, transform_gens):
        tfms_equivariant = []
        tfms_invariant = []
        for tfm in transform_gens:
            if isinstance(tfm, TransformGen):
                tfm_kind = tfm._base_transform
            else:
                tfm_kind = type(tfm)
            assert issubclass(tfm_kind, Transform)

            if issubclass(tfm_kind, GeometricMixin):
                tfms_equivariant.append(tfm)
            else:
                tfms_invariant.append(tfm)
        return tfms_equivariant, tfms_invariant

    def _permute_data(self, *args, spatial_last: bool = False):
        out = []
        if spatial_last:
            for x in args:
                dims = (0,) + tuple(range(3, x.ndim)) + (1, 2)
                out.append(x.permute(dims))
        else:
            for x in args:
                dims = (0, x.ndim - 2, x.ndim - 1) + tuple(range(1, x.ndim - 2))
                out.append(x.permute(dims))
        return out[0] if len(out) == 1 else tuple(out)

    def _apply_te(self, tfms_equivariant, image, target, maps):
        """Apply equivariant transforms.

        These transforms affect both the input and the target.
        """
        tfms = []
        for g in tfms_equivariant:
            tfm: Transform = g.get_transform(image) if isinstance(g, TransformGen) else g
            if isinstance(tfm, NoOpTransform):
                continue
            image = tfm.apply_image(image)
            if target is not None:
                target = tfm.apply_image(target)
            if maps is not None and self.aug_sensitivity_maps:
                maps = tfm.apply_maps(maps)
            tfms.append(tfm)
        return image, target, maps, TransformList(tfms, ignore_no_op=True)

    def _apply_ti(self, tfms_invariant, kspace):
        """Apply invariant transforms.

        These transforms affect only the input, not the target.
        """
        tfms = []
        for g in tfms_invariant:
            tfm: Transform = g.get_transform(kspace) if isinstance(g, TransformGen) else g
            if isinstance(tfm, NoOpTransform):
                continue
            kspace = tfm.apply_kspace(kspace)
            tfms.append(tfm)
        return kspace, TransformList(tfms, ignore_no_op=True)

    def reset(self):
        for g in self.tfms_or_gens:
            if isinstance(g, TransformGen):
                g.reset()

    def to(self, device):
        tfms = [tfm for tfm in self.tfms_or_gens if isinstance(tfm, DeviceMixin)]
        for t in tfms:
            t.to(device)
        return self

    @classmethod
    def from_cfg(cls, cfg, aug_kind, seed=None, device=None, **kwargs):
        mri_tfm_cfg = None
        assert aug_kind in ("aug_train", "consistency")
        if aug_kind == "aug_train":
            mri_tfm_cfg = cfg.AUG_TRAIN.MRI_RECON
        elif aug_kind == "consistency":
            mri_tfm_cfg = cfg.MODEL.CONSISTENCY.AUG.MRI_RECON

        if seed is None and env.is_repro():
            seed = cfg.SEED

        tfms_or_gens = build_transforms(cfg, mri_tfm_cfg.TRANSFORMS, seed=seed, **kwargs)
        scheduler_p = dict(mri_tfm_cfg.SCHEDULER_P)
        ignore_scheduler = scheduler_p.pop("IGNORE", False)
        if len(scheduler_p) and not ignore_scheduler:
            scheduler_p["params"] = ["p"]
            tfms = [tfm for tfm in tfms_or_gens if isinstance(tfm, TransformGen)]
            for tfm in tfms:
                scheduler = build_scheduler(cfg, scheduler_p, tfm)
                tfm.register_schedulers([scheduler])

        if aug_kind in ("aug_train",) and cfg.DATALOADER.NUM_WORKERS > 0:
            func = build_iter_func(cfg.SOLVER.TRAIN_BATCH_SIZE, cfg.DATALOADER.NUM_WORKERS)
            tfms = [tfm for tfm in tfms_or_gens if isinstance(tfm, SchedulableMixin)]
            for tfm in tfms:
                for s in tfm._schedulers:
                    s._iter_fn = func

        return cls(
            tfms_or_gens,
            aug_sensitivity_maps=mri_tfm_cfg.AUG_SENSITIVITY_MAPS,
            seed=seed,
            device=device,
        )



from .evaluator import DatasetEvaluator, DatasetEvaluators, inference_on_dataset  # noqa: F401
from .recon_evaluation import ReconEvaluator  # noqa: F401
from .testing import check_consistency, find_weights, print_csv_format, verify_results  # noqa: F401



"""
Utilities for doing complex-valued operations.
"""
import numpy as np
import torch

from meddlr.utils.deprecated import deprecated
from meddlr.utils.env import supports_cplx_tensor

__all__ = [
    "is_complex",
    "is_complex_as_real",
    "conj",
    "mul",
    "abs",
    "angle",
    "real",
    "imag",
    "from_polar",
    "channels_first",
    "channels_last",
    "get_mask",
    "matmul",
    "power_method",
    "svd",
    "to_numpy",
    "to_tensor",
    "rss",
    "center_crop",
]


def is_complex(x):
    """Returns if ``x`` is a complex-tensor.

    This function is a wrapper around torch.is_complex() for PyTorch<1.7.
    torch < 1.7 does not have the ``torch.is_complex`` directive, so
    we can't call it for older PyTorch versions.

    Args:
        x (torch.Tensor): A tensor.

    Returns:
        bool: ``True`` if complex tensors are supported and the tensor is complex.
    """
    return supports_cplx_tensor() and torch.is_complex(x)


def is_complex_as_real(x):
    """
    Returns ``True`` if the tensor follows the real-view
    convention for complex numbers.

    The real-view of a complex tensor has the shape [..., 2].

    Args:
        x (torch.Tensor): A tensor.

    Returns:
        bool: ``True`` if the tensor follows the real-view convention
            for complex numbers.

    Note:
        We recommend using complex tensors instead of the real-view
        convention. This function cannot interpret if the last dimension
        has a size of ``2`` because it is the real-imaginary channel or
        for some other reason.
    """
    return x.size(-1) == 2


def conj(x):
    """
    Computes the complex conjugate of complex-valued input tensor (x).

    ``conj(a + ib)`` = :math:`\\bar{a + ib} = a - ib`

    Args:
        x (torch.Tensor): A tensor.

    Returns:
        torch.Tensor: The conjugate.
    """
    assert is_complex_as_real(x) or is_complex(x)
    if is_complex(x):
        return x.conj()
    else:
        real = x[..., 0]
        imag = x[..., 1]
        return torch.stack((real, -1.0 * imag), dim=-1)


def mul(x, y):
    """
    Multiplies two complex-valued tensors x and y.

    :math:`z = (a + ib) * (c + id) = (ac - bd) + i(ad + bc)`

    Args:
        x (torch.Tensor): A tensor.
        y (torch.Tensor): A tensor.

    Returns:
        torch.Tensor: The matrix multiplication.
    """
    # assert x.size(-1) == 2
    # assert y.size(-1) == 2
    #
    # a = x[..., 0]
    # b = x[..., 1]
    # c = y[..., 0]
    # d = y[..., 1]
    #
    # real = a * c - b * d
    # imag = a * d + b * c

    # return torch.stack((real, imag), dim=-1)

    assert is_complex_as_real(x) or is_complex(x)
    assert is_complex_as_real(y) or is_complex(y)
    if is_complex(x):
        return x * y
    else:
        # note: using select() makes sure that another copy is not made.
        # real = a*c - b*d
        real = x.select(-1, 0) * y.select(-1, 0)  # a*c
        real -= x.select(-1, 1) * y.select(-1, 1)  # b*d
        # imag = a*d + b*c
        imag = x.select(-1, 0) * y.select(-1, 1)  # a*d
        imag += x.select(-1, 1) * y.select(-1, 0)  # b*c
        return torch.stack((real, imag), dim=-1)


def abs(x):
    """
    Computes the absolute value (magnitude) of a complex-valued input tensor (x).

    Args:
        x (torch.Tensor): A tensor.

    Returns:
        torch.Tensor: The magnitude tensor.
    """
    assert is_complex_as_real(x) or is_complex(x)
    if is_complex(x):
        return x.abs()
    else:
        return (x ** 2).sum(dim=-1).sqrt()


def angle(x, eps=1e-11):
    """
    Computes the phase of a complex-valued input tensor (x).

    Args:
        x (torch.Tensor): A tensor.

    Returns:
        torch.Tensor: The angle tensor.
    """
    assert is_complex_as_real(x) or is_complex(x)
    if is_complex(x):
        return x.angle()
    else:
        return torch.atan(x[..., 1] / (x[..., 0] + eps))


def real(x):
    """
    Gets real component of complex tensor.
    """
    assert is_complex_as_real(x) or is_complex(x)
    if is_complex(x):
        return x.real
    else:
        return x[..., 0]


def imag(x):
    """
    Gets imaginary component of complex tensor.
    """
    assert is_complex_as_real(x) or is_complex(x)
    if is_complex(x):
        return x.imag
    else:
        return x[..., 1]


def from_polar(magnitude, phase, return_cplx: bool = False):
    """
    Computes real and imaginary values from polar representation.
    """
    if return_cplx:
        if not supports_cplx_tensor():
            raise RuntimeError(f"torch {torch.__version__} does not support complex tensors")
        return torch.polar(magnitude, phase)
    real = magnitude * torch.cos(phase)
    imag = magnitude * torch.sin(phase)
    return torch.stack((real, imag), dim=-1)


def channels_first(x: torch.Tensor):
    """Permute complex-valued ``x`` to channels-first convention.

    For complex values, there are two potential conventions:

    1. ``x`` is complex-valued: ``(B,...,C)`` -> ``(B, C, ...)``.
    2. The real and imaginary components are stored in the last dimension.
       ``(B,...,C,2)`` -> ``(B, C, ..., 2)``.

    Args:
        x (torch.Tensor): A complex-valued tensor of shape ``(B,...,C)``
            or a real-valued tensor of shape ``(B,...,C,2)``.

    Returns:
        torch.Tensor: A channels-first tensor. If ``x`` is complex,
            this will also be complex. If ``x`` is the real-view of
            a complex tensor, this will also be the real view.
    """
    assert is_complex_as_real(x) or is_complex(x)
    if is_complex(x):
        return x.permute((0, x.ndim - 1) + tuple(range(1, x.ndim - 1)))
    else:
        return x.permute((0, x.ndim - 2) + tuple(range(1, x.ndim - 2)) + (x.ndim - 1,))


@deprecated(
    reason="Renamed to channels_first",
    vremove="v0.1.0",
    replacement="meddlr.ops.complex.channels_first",
)
def channel_first(x: torch.Tensor):
    """Deprecated alias for :func:`channels_first`."""
    return channels_first(x)


def channels_last(x: torch.Tensor):
    """Permute complex-valued ``x`` to channels-last convention.

    Args:
        x (torch.Tensor): A tensor of shape [B,C,H,W,...] or [B,C,H,W,...,2].

    Returns:
        torch.Tensor: A tensor of shape [B,H,W,...,C] or [B,H,W,...,C,2].
    """
    assert is_complex_as_real(x) or is_complex(x)
    if is_complex(x):
        return x.permute((0,) + tuple(range(2, x.ndim) + (1,)))
    else:
        order = (0,) + tuple(range(2, x.ndim - 2)) + (1, x.ndim - 1)
        return x.permute(order)


def get_mask(x, eps=1e-11):
    """Returns a binary mask for where ``x`` is nonzero with ``eps`` tolerance.

      - 0, if both real and imaginary components are zero.
      - 1, if either real and imaginary components are non-zero.

    Args:
        x (torch.Tensor): A complex-valued tensor.
        eps (float): Tolerance for zer0-value.

    Returns:
        torch.Tensor: A binary mask of shape ``x.shape``.
    """
    unsqueeze = True
    if is_complex(x):
        unsqueeze = False
        x = torch.view_as_real(x)
    assert x.size(-1) == 2
    absx = abs(x)  # squashes last dimension
    mask = torch.where(absx > eps, torch.ones_like(absx), torch.zeros_like(absx))
    if unsqueeze:
        mask = mask.unsqueeze(-1)
    return mask


def matmul(X, Y):
    """
    Computes complex-valued matrix product of X and Y.
    """
    assert is_complex_as_real(X) or is_complex(X)
    assert is_complex_as_real(Y) or is_complex(Y)
    if is_complex(X):
        return torch.matmul(X, Y)
    else:
        A = X[..., 0]
        B = X[..., 1]
        C = Y[..., 0]
        D = Y[..., 1]
        real = torch.matmul(A, C) - torch.matmul(B, D)
        imag = torch.matmul(A, D) + torch.matmul(B, C)
        return torch.stack((real, imag), dim=-1)


def power_method(X, num_iter=10, eps=1e-6):
    """
    Iteratively computes first singular value of X using power method.
    """
    if is_complex_as_real(X) or is_complex(X):
        X = torch.view_as_real(X)
    assert X.size(-1) == 2

    # get data dimensions
    batch_size, m, n, _ = X.shape

    XhX = matmul(conj(X).permute(0, 2, 1, 3), X)

    # initialize random eigenvector
    if XhX.is_cuda:
        v = torch.cuda.FloatTensor(batch_size, n, 1, 2).uniform_()
    else:
        v = torch.FloatTensor(batch_size, n, 1, 2).uniform_()
    # v = torch.rand(batch_size, n, 1, 2).to(X.device) # slow way

    for _i in range(num_iter):
        v = matmul(XhX, v)
        eigenvals = (abs(v) ** 2).sum(1).sqrt()
        v = v / (eigenvals.reshape(batch_size, 1, 1, 1) + eps)

    return eigenvals.reshape(batch_size)


def svd(X, compute_uv=True):
    """
    Computes singular value decomposition of batch of complex-valued matrices.

    Args:
        matrix (torch.Tensor): batch of complex-valued 2D matrices
            [batch, m, n, 2]
    Returns:
        U, S, V (tuple)
    """
    if is_complex_as_real(X) or is_complex(X):
        X = torch.view_as_real(X)
    assert X.size(-1) == 2

    # Get data dimensions
    batch_size, m, n, _ = X.shape

    # Allocate block-wise matrix
    # (otherwise, need to allocate new arrays three times)
    if X.is_cuda:
        Xb = torch.cuda.FloatTensor(batch_size, 2 * m, 2 * n).fill_(0)
    else:
        Xb = torch.FloatTensor(batch_size, 2 * m, 2 * n).fill_(0)

    # Construct real-valued block matrix
    # Xb = [X.real, X.imag; -X.imag, X.real]
    Xb[:, :m, :n] = X[..., 0]
    Xb[:, :m, n:] = X[..., 1]
    Xb[:, m:, :n] = -X[..., 1]
    Xb[:, m:, n:] = X[..., 0]

    # Perform real-valued SVD
    U, S, V = torch.svd(Xb, compute_uv=compute_uv)

    # Slice U, S, V appropriately
    S = S[:, ::2]
    U = torch.stack((U[:, :m, ::2], -U[:, m:, ::2]), dim=3)
    V = torch.stack((V[:, :n, ::2], -V[:, n:, ::2]), dim=3)

    return U, S, V


def to_numpy(x: torch.Tensor):
    """
    Convert real-valued PyTorch tensor to complex-valued numpy array.
    """
    assert is_complex_as_real(x) or is_complex(x)
    if is_complex(x):
        return x.clone().numpy()  # previously returned copy
    else:
        x = x.numpy()
        return x[..., 0] + 1j * x[..., 1]


def to_tensor(x: np.ndarray):
    """
    Convert complex-valued numpy array to real-valued PyTorch tensor.
    """
    if not supports_cplx_tensor():
        x = np.stack((x.real, x.imag), axis=-1)
    return torch.from_numpy(x)


def rss(x: torch.Tensor, dim: int = 0) -> torch.Tensor:
    """
    Compute the root-sum-of-squares (RSS) for complex inputs.
    RSS is computed assuming that dim is the coil dimension.

    Args:
        data: The complex-valued input tensor.
        dim: The dimensions along which to apply the RSS transform.

    Returns:
        torch.Tensor: The RSS value.
    """
    return torch.sqrt((abs(x) ** 2).sum(dim))


root_sum_of_squares = rss


def center_crop(x: torch.Tensor, shape, channels_last: bool = False):
    """
    Apply a center crop to the input image or batch of complex images.

    Args:
        data (torch.Tensor): The complex input tensor to be center cropped.
        shape (int, int): The output shape. The shape should be smaller than the
            corresponding dimensions of data.
        channels_last (bool, optional): If ``True``, crop dimensions ``range(1, 1+len(shape))``.
            If ``False``, apply to last non-real/imaginary channel dimensions.

    Returns:
        torch.Tensor: The center cropped image.
    """
    if channels_last:
        dims = range(1, 1 + len(shape))
    elif not is_complex(x) and is_complex_as_real(x):
        dims = range(-1 - len(shape), -1)
    else:
        dims = range(-len(shape), 0)

    x_shape = tuple(x.shape[d] for d in dims)
    assert all(0 < shape[idx] <= x_shape[idx] for idx in range(len(shape)))

    sl = [slice(None) for _ in range(x.ndim)]
    for d, shp, x_shp in zip(dims, shape, x_shape):
        start = (x_shp - shp) // 2
        end = start + shp
        sl[d] = slice(start, end)
    return x[sl]
